import collections
import math

import numpy as np
from scipy import sparse
from scipy.linalg import block_diag

from qibo import __version__
from qibo.backends import einsum_utils
from qibo.backends.abstract import Backend
from qibo.backends.npmatrices import NumpyMatrices
from qibo.config import log, raise_error
from qibo.result import CircuitResult, MeasurementOutcomes, QuantumState


class NumpyBackend(Backend):
    def __init__(self):
        super().__init__()
        self.np = np
        self.name = "numpy"
        self.matrices = NumpyMatrices(self.dtype)
        self.tensor_types = np.ndarray
        self.versions = {"qibo": __version__, "numpy": self.np.__version__}
        self.numeric_types = (
            int,
            float,
            complex,
            np.int32,
            np.int64,
            np.float32,
            np.float64,
            np.complex64,
            np.complex128,
        )

    def set_precision(self, precision):
        if precision != self.precision:
            if precision == "single":
                self.precision = precision
                self.dtype = "complex64"
            elif precision == "double":
                self.precision = precision
                self.dtype = "complex128"
            else:
                raise_error(ValueError, f"Unknown precision {precision}.")
            if self.matrices:
                self.matrices = self.matrices.__class__(self.dtype)

    def set_device(self, device):
        if device != "/CPU:0":
            raise_error(
                ValueError, f"Device {device} is not available for {self} backend."
            )

    def set_threads(self, nthreads):
        if nthreads > 1:
            raise_error(ValueError, "numpy does not support more than one thread.")

    def cast(self, x, dtype=None, copy=False):
        if dtype is None:
            dtype = self.dtype
        if isinstance(x, self.tensor_types):
            return x.astype(dtype, copy=copy)
        elif self.issparse(x):
            return x.astype(dtype, copy=copy)
        return np.array(x, dtype=dtype, copy=copy)

    def issparse(self, x):
        from scipy import sparse

        return sparse.issparse(x)

    def to_numpy(self, x):
        if self.issparse(x):
            return x.toarray()
        return x

    def compile(self, func):
        return func

    def zero_state(self, nqubits):
        state = self.np.zeros(2**nqubits, dtype=self.dtype)
        state[0] = 1
        return state

    def zero_density_matrix(self, nqubits):
        state = self.np.zeros(2 * (2**nqubits,), dtype=self.dtype)
        state[0, 0] = 1
        return state

    def identity_density_matrix(self, nqubits, normalize: bool = True):
        state = self.np.eye(2**nqubits, dtype=self.dtype)
        if normalize is True:
            state /= 2**nqubits
        return state

    def plus_state(self, nqubits):
        state = self.np.ones(2**nqubits, dtype=self.dtype)
        state /= math.sqrt(2**nqubits)
        return state

    def plus_density_matrix(self, nqubits):
        state = self.np.ones(2 * (2**nqubits,), dtype=self.dtype)
        state /= 2**nqubits
        return state

    def matrix(self, gate):
        """Convert a gate to its matrix representation in the computational basis."""
        name = gate.__class__.__name__
        _matrix = getattr(self.matrices, name)
        if callable(_matrix):
            _matrix = _matrix(2 ** len(gate.target_qubits))
        return self.cast(_matrix, dtype=_matrix.dtype)

    def matrix_parametrized(self, gate):
        """Convert a parametrized gate to its matrix representation in the computational basis."""
        name = gate.__class__.__name__
        _matrix = getattr(self.matrices, name)(*gate.parameters)
        return self.cast(_matrix, dtype=_matrix.dtype)

    def matrix_fused(self, fgate):
        rank = len(fgate.target_qubits)
        matrix = sparse.eye(2**rank)

        for gate in fgate.gates:
            # transfer gate matrix to numpy as it is more efficient for
            # small tensor calculations
            # explicit to_numpy see https://github.com/qiboteam/qibo/issues/928
            gmatrix = self.to_numpy(gate.matrix(self))
            # add controls if controls were instantiated using
            # the ``Gate.controlled_by`` method
            num_controls = len(gate.control_qubits)
            if num_controls > 0:
                gmatrix = block_diag(
                    np.eye(2 ** len(gate.qubits) - len(gmatrix)), gmatrix
                )
            # Kronecker product with identity is needed to make the
            # original matrix have shape (2**rank x 2**rank)
            eye = np.eye(2 ** (rank - len(gate.qubits)))
            gmatrix = np.kron(gmatrix, eye)
            # Transpose the new matrix indices so that it targets the
            # target qubits of the original gate
            original_shape = gmatrix.shape
            gmatrix = np.reshape(gmatrix, 2 * rank * (2,))
            qubits = list(gate.qubits)
            indices = qubits + [q for q in fgate.target_qubits if q not in qubits]
            indices = np.argsort(indices)
            transpose_indices = list(indices)
            transpose_indices.extend(indices + rank)
            gmatrix = np.transpose(gmatrix, transpose_indices)
            gmatrix = np.reshape(gmatrix, original_shape)
            # fuse the individual gate matrix to the total ``FusedGate`` matrix
            # we are using sparse matrices to improve perfomances
            matrix = sparse.csr_matrix(gmatrix).dot(matrix)

        return self.cast(matrix.toarray())

    def apply_gate(self, gate, state, nqubits):
        state = self.cast(state)
        state = self.np.reshape(state, nqubits * (2,))
        matrix = gate.matrix(self)
        if gate.is_controlled_by:
            matrix = self.np.reshape(matrix, 2 * len(gate.target_qubits) * (2,))
            ncontrol = len(gate.control_qubits)
            nactive = nqubits - ncontrol
            order, targets = einsum_utils.control_order(gate, nqubits)
            state = self.np.transpose(state, order)
            # Apply `einsum` only to the part of the state where all controls
            # are active. This should be `state[-1]`
            state = self.np.reshape(state, (2**ncontrol,) + nactive * (2,))
            opstring = einsum_utils.apply_gate_string(targets, nactive)
            updates = self.np.einsum(opstring, state[-1], matrix)
            # Concatenate the updated part of the state `updates` with the
            # part of of the state that remained unaffected `state[:-1]`.
            state = self.np.concatenate([state[:-1], updates[None]], axis=0)
            state = self.np.reshape(state, nqubits * (2,))
            # Put qubit indices back to their proper places
            state = self.np.transpose(state, einsum_utils.reverse_order(order))
        else:
            matrix = self.np.reshape(matrix, 2 * len(gate.qubits) * (2,))
            opstring = einsum_utils.apply_gate_string(gate.qubits, nqubits)
            state = self.np.einsum(opstring, state, matrix)
        return self.np.reshape(state, (2**nqubits,))

    def apply_gate_density_matrix(self, gate, state, nqubits):
        state = self.cast(state)
        state = self.np.reshape(state, 2 * nqubits * (2,))
        matrix = gate.matrix(self)
        if gate.is_controlled_by:
            matrix = self.np.reshape(matrix, 2 * len(gate.target_qubits) * (2,))
            matrixc = self.np.conj(matrix)
            ncontrol = len(gate.control_qubits)
            nactive = nqubits - ncontrol
            n = 2**ncontrol

            order, targets = einsum_utils.control_order_density_matrix(gate, nqubits)
            state = self.np.transpose(state, order)
            state = self.np.reshape(state, 2 * (n,) + 2 * nactive * (2,))

            leftc, rightc = einsum_utils.apply_gate_density_matrix_controlled_string(
                targets, nactive
            )
            state01 = state[: n - 1, n - 1]
            state01 = self.np.einsum(rightc, state01, matrixc)
            state10 = state[n - 1, : n - 1]
            state10 = self.np.einsum(leftc, state10, matrix)

            left, right = einsum_utils.apply_gate_density_matrix_string(
                targets, nactive
            )
            state11 = state[n - 1, n - 1]
            state11 = self.np.einsum(right, state11, matrixc)
            state11 = self.np.einsum(left, state11, matrix)

            state00 = state[range(n - 1)]
            state00 = state00[:, range(n - 1)]
            state01 = self.np.concatenate([state00, state01[:, None]], axis=1)
            state10 = self.np.concatenate([state10, state11[None]], axis=0)
            state = self.np.concatenate([state01, state10[None]], axis=0)
            state = self.np.reshape(state, 2 * nqubits * (2,))
            state = self.np.transpose(state, einsum_utils.reverse_order(order))
        else:
            matrix = self.np.reshape(matrix, 2 * len(gate.qubits) * (2,))
            matrixc = self.np.conj(matrix)
            left, right = einsum_utils.apply_gate_density_matrix_string(
                gate.qubits, nqubits
            )
            state = self.np.einsum(right, state, matrixc)
            state = self.np.einsum(left, state, matrix)
        return self.np.reshape(state, 2 * (2**nqubits,))

    def apply_gate_half_density_matrix(self, gate, state, nqubits):
        state = self.cast(state)
        state = self.np.reshape(state, 2 * nqubits * (2,))
        matrix = gate.matrix(self)
        if gate.is_controlled_by:  # pragma: no cover
            raise_error(
                NotImplementedError,
                "Gate density matrix half call is "
                "not implemented for ``controlled_by``"
                "gates.",
            )
        else:
            matrix = self.np.reshape(matrix, 2 * len(gate.qubits) * (2,))
            left, _ = einsum_utils.apply_gate_density_matrix_string(
                gate.qubits, nqubits
            )
            state = self.np.einsum(left, state, matrix)
        return self.np.reshape(state, 2 * (2**nqubits,))

    def apply_channel(self, channel, state, nqubits):
        probabilities = channel.coefficients + (1 - np.sum(channel.coefficients),)
        index = self.sample_shots(probabilities, 1)[0]
        if index != len(channel.gates):
            gate = channel.gates[index]
            state = self.apply_gate(gate, state, nqubits)
        return state

    def apply_channel_density_matrix(self, channel, state, nqubits):
        state = self.cast(state)
        new_state = (1 - channel.coefficient_sum) * state
        for coeff, gate in zip(channel.coefficients, channel.gates):
            new_state += coeff * self.apply_gate_density_matrix(gate, state, nqubits)
        return new_state

    def _append_zeros(self, state, qubits, results):
        """Helper method for collapse."""
        for q, r in zip(qubits, results):
            state = self.np.expand_dims(state, q)
            state = (
                self.np.concatenate([self.np.zeros_like(state), state], q)
                if r == 1
                else self.np.concatenate([state, self.np.zeros_like(state)], q)
            )
        return state

    def collapse_state(self, state, qubits, shot, nqubits, normalize=True):
        state = self.cast(state)
        shape = state.shape
        binshot = self.samples_to_binary(shot, len(qubits))[0]
        state = self.np.reshape(state, nqubits * (2,))
        order = list(qubits) + [q for q in range(nqubits) if q not in qubits]
        state = self.np.transpose(state, order)
        subshape = (2 ** len(qubits),) + (nqubits - len(qubits)) * (2,)
        state = self.np.reshape(state, subshape)[int(shot)]
        if normalize:
            norm = self.np.sqrt(self.np.sum(self.np.abs(state) ** 2))
            state = state / norm
        state = self._append_zeros(state, qubits, binshot)
        return self.np.reshape(state, shape)

    def collapse_density_matrix(self, state, qubits, shot, nqubits, normalize=True):
        state = self.cast(state)
        shape = state.shape
        binshot = list(self.samples_to_binary(shot, len(qubits))[0])
        order = list(qubits) + [q + nqubits for q in qubits]
        order.extend(q for q in range(nqubits) if q not in qubits)
        order.extend(q + nqubits for q in range(nqubits) if q not in qubits)
        state = self.np.reshape(state, 2 * nqubits * (2,))
        state = self.np.transpose(state, order)
        subshape = 2 * (2 ** len(qubits),) + 2 * (nqubits - len(qubits)) * (2,)
        state = self.np.reshape(state, subshape)[int(shot), int(shot)]
        n = 2 ** (len(state.shape) // 2)
        if normalize:
            norm = self.np.trace(self.np.reshape(state, (n, n)))
            state = state / norm
        qubits = qubits + [q + nqubits for q in qubits]
        state = self._append_zeros(state, qubits, 2 * binshot)
        return self.np.reshape(state, shape)

    def reset_error_density_matrix(self, gate, state, nqubits):
        from qibo.gates import X

        state = self.cast(state)
        shape = state.shape
        q = gate.target_qubits[0]
        p_0, p_1 = gate.init_kwargs["p_0"], gate.init_kwargs["p_1"]
        trace = self.partial_trace_density_matrix(state, (q,), nqubits)
        trace = self.np.reshape(trace, 2 * (nqubits - 1) * (2,))
        zero = self.zero_density_matrix(1)
        zero = self.np.tensordot(trace, zero, 0)
        order = list(range(2 * nqubits - 2))
        order.insert(q, 2 * nqubits - 2)
        order.insert(q + nqubits, 2 * nqubits - 1)
        zero = self.np.reshape(self.np.transpose(zero, order), shape)
        state = (1 - p_0 - p_1) * state + p_0 * zero
        return state + p_1 * self.apply_gate_density_matrix(X(q), zero, nqubits)

    def thermal_error_density_matrix(self, gate, state, nqubits):
        state = self.cast(state)
        shape = state.shape
        state = self.apply_gate(gate, state.ravel(), 2 * nqubits)
        return self.np.reshape(state, shape)

    def depolarizing_error_density_matrix(self, gate, state, nqubits):
        state = self.cast(state)
        shape = state.shape
        q = gate.target_qubits
        lam = gate.init_kwargs["lam"]
        trace = self.partial_trace_density_matrix(state, q, nqubits)
        trace = self.np.reshape(trace, 2 * (nqubits - len(q)) * (2,))
        identity = self.identity_density_matrix(len(q))
        identity = self.np.reshape(identity, 2 * len(q) * (2,))
        identity = self.np.tensordot(trace, identity, 0)
        qubits = list(range(nqubits))
        for j in q:
            qubits.pop(qubits.index(j))
        qubits.sort()
        qubits += list(q)
        qubit_1 = list(range(nqubits - len(q))) + list(
            range(2 * (nqubits - len(q)), 2 * nqubits - len(q))
        )
        qubit_2 = list(range(nqubits - len(q), 2 * (nqubits - len(q)))) + list(
            range(2 * nqubits - len(q), 2 * nqubits)
        )
        qs = [qubit_1, qubit_2]
        order = []
        for qj in qs:
            qj = [qj[qubits.index(i)] for i in range(len(qubits))]
            order += qj
        identity = self.np.reshape(self.np.transpose(identity, order), shape)
        state = (1 - lam) * state + lam * identity
        return state

    def execute_circuit(self, circuit, initial_state=None, nshots=1000):

        if isinstance(initial_state, type(circuit)):
            if not initial_state.density_matrix == circuit.density_matrix:
                raise_error(
                    ValueError,
                    f"""Cannot set circuit with density_matrix {initial_state.density_matrix} as
                      initial state for circuit with density_matrix {circuit.density_matrix}.""",
                )
            elif (
                not initial_state.accelerators == circuit.accelerators
            ):  # pragma: no cover
                raise_error(
                    ValueError,
                    f"""Cannot set circuit with accelerators {initial_state.density_matrix} as
                      initial state for circuit with accelerators {circuit.density_matrix}.""",
                )
            else:
                return self.execute_circuit(initial_state + circuit, None, nshots)
        elif initial_state is not None:
            initial_state = self.cast(initial_state)
            valid_shape = (
                2 * (2**circuit.nqubits,)
                if circuit.density_matrix
                else (2**circuit.nqubits,)
            )
            if tuple(initial_state.shape) != valid_shape:
                raise_error(
                    ValueError,
                    f"Given initial state has shape {initial_state.shape} instead of "
                    f"the expected {valid_shape}.",
                )

        if circuit.repeated_execution:
            if circuit.measurements or circuit.has_collapse:
                return self.execute_circuit_repeated(circuit, nshots, initial_state)
            else:
                raise RuntimeError(
                    "Attempting to perform noisy simulation with `density_matrix=False` and no Measurement gate in the Circuit. If you wish to retrieve the statistics of the outcomes please include measurements in the circuit, otherwise set `density_matrix=True` to recover the final state."
                )

        if circuit.accelerators:  # pragma: no cover
            return self.execute_distributed_circuit(circuit, initial_state, nshots)

        try:
            nqubits = circuit.nqubits

            if circuit.density_matrix:
                if initial_state is None:
                    state = self.zero_density_matrix(nqubits)
                else:
                    # cast to proper complex type
                    state = self.cast(initial_state)

                for gate in circuit.queue:
                    state = gate.apply_density_matrix(self, state, nqubits)

            else:
                if initial_state is None:
                    state = self.zero_state(nqubits)
                else:
                    # cast to proper complex type
                    state = self.cast(initial_state)

                for gate in circuit.queue:
                    state = gate.apply(self, state, nqubits)

            if circuit.has_unitary_channel:
                # here we necessarily have `density_matrix=True`, otherwise
                # execute_circuit_repeated would have been called
                if circuit.measurements:
                    circuit._final_state = CircuitResult(
                        state, circuit.measurements, backend=self, nshots=nshots
                    )
                    return circuit._final_state
                else:
                    circuit._final_state = QuantumState(state, backend=self)
                    return circuit._final_state

            else:
                if circuit.measurements:
                    circuit._final_state = CircuitResult(
                        state, circuit.measurements, backend=self, nshots=nshots
                    )
                    return circuit._final_state
                else:
                    circuit._final_state = QuantumState(state, backend=self)
                    return circuit._final_state

        except self.oom_error:
            raise_error(
                RuntimeError,
                f"State does not fit in {self.device} memory."
                "Please switch the execution device to a "
                "different one using ``qibo.set_device``.",
            )

    def execute_circuits(
        self, circuits, initial_states=None, nshots=1000, processes=None
    ):
        from qibo.parallel import parallel_circuits_execution

        return parallel_circuits_execution(
            circuits, initial_states, nshots, processes, backend=self
        )

    def execute_circuit_repeated(self, circuit, nshots, initial_state=None):
        """
        Execute the circuit `nshots` times to retrieve probabilities, frequencies
        and samples. Note that this method is called only if a unitary channel
        is present in the circuit (i.e. noisy simulation) and `density_matrix=False`, or
        if some collapsing measurement is performed.
        """

        if (
            circuit.has_collapse
            and not circuit.measurements
            and not circuit.density_matrix
        ):
            raise RuntimeError(
                "The circuit contains only collapsing measurements (`collapse=True`) but `density_matrix=False`. Please set `density_matrix=True` to retrieve the final state after execution."
            )

        results, final_states = [], []
        nqubits = circuit.nqubits

        if not circuit.density_matrix:
            samples = []
            target_qubits = [
                measurement.target_qubits for measurement in circuit.measurements
            ]
            target_qubits = sum(target_qubits, tuple())

        for _ in range(nshots):
            if circuit.density_matrix:
                if initial_state is None:
                    state = self.zero_density_matrix(nqubits)
                else:
                    state = self.cast(initial_state, copy=True)

                for gate in circuit.queue:
                    if gate.symbolic_parameters:
                        gate.substitute_symbols()
                    state = gate.apply_density_matrix(self, state, nqubits)
            else:
                if circuit.accelerators:  # pragma: no cover
                    # pylint: disable=E1111
                    state = self.execute_distributed_circuit(circuit, initial_state)
                else:
                    if initial_state is None:
                        state = self.zero_state(nqubits)
                    else:
                        state = self.cast(initial_state, copy=True)

                    for gate in circuit.queue:
                        if gate.symbolic_parameters:
                            gate.substitute_symbols()
                        state = gate.apply(self, state, nqubits)

            if circuit.density_matrix:
                final_states.append(state)
            if circuit.measurements:
                result = CircuitResult(
                    state, circuit.measurements, backend=self, nshots=1
                )
                sample = result.samples()[0]
                results.append(sample)
                if not circuit.density_matrix:
                    samples.append("".join([str(int(s)) for s in sample]))
                for gate in circuit.measurements:
                    gate.result.reset()

        if circuit.density_matrix:  # this implies also it has_collapse
            assert circuit.has_collapse
            final_state = self.cast(np.mean(self.to_numpy(final_states), 0))
            if circuit.measurements:
                final_result = CircuitResult(
                    final_state,
                    circuit.measurements,
                    backend=self,
                    samples=self.aggregate_shots(results),
                    nshots=nshots,
                )
            else:
                final_result = QuantumState(final_state, backend=self)
            circuit._final_state = final_result
            return final_result
        else:
            final_result = MeasurementOutcomes(
                circuit.measurements,
                backend=self,
                samples=self.aggregate_shots(results),
                nshots=nshots,
            )
            final_result._repeated_execution_frequencies = self.calculate_frequencies(
                samples
            )
            circuit._final_state = final_result
            return final_result

    def execute_distributed_circuit(self, circuit, initial_state=None, nshots=None):
        raise_error(
            NotImplementedError, f"{self} does not support distributed execution."
        )

    def calculate_symbolic(
        self, state, nqubits, decimals=5, cutoff=1e-10, max_terms=20
    ):
        state = self.to_numpy(state)
        terms = []
        for i in np.nonzero(state)[0]:
            b = bin(i)[2:].zfill(nqubits)
            if np.abs(state[i]) >= cutoff:
                x = np.round(state[i], decimals)
                terms.append(f"{x}|{b}>")
            if len(terms) >= max_terms:
                terms.append("...")
                return terms
        return terms

    def calculate_symbolic_density_matrix(
        self, state, nqubits, decimals=5, cutoff=1e-10, max_terms=20
    ):
        state = self.to_numpy(state)
        terms = []
        indi, indj = np.nonzero(state)
        for i, j in zip(indi, indj):
            bi = bin(i)[2:].zfill(nqubits)
            bj = bin(j)[2:].zfill(nqubits)
            if np.abs(state[i, j]) >= cutoff:
                x = np.round(state[i, j], decimals)
                terms.append(f"{x}|{bi}><{bj}|")
            if len(terms) >= max_terms:
                terms.append("...")
                return terms
        return terms

    def _order_probabilities(self, probs, qubits, nqubits):
        """Arrange probabilities according to the given ``qubits`` ordering."""
        unmeasured, reduced = [], {}
        for i in range(nqubits):
            if i in qubits:
                reduced[i] = i - len(unmeasured)
            else:
                unmeasured.append(i)
        return self.np.transpose(probs, [reduced.get(i) for i in qubits])

    def calculate_probabilities(self, state, qubits, nqubits):
        rtype = self.np.real(state).dtype
        unmeasured_qubits = tuple(i for i in range(nqubits) if i not in qubits)
        state = self.np.reshape(self.np.abs(state) ** 2, nqubits * (2,))
        probs = self.np.sum(self.cast(state, dtype=rtype), axis=unmeasured_qubits)
        return self._order_probabilities(probs, qubits, nqubits).ravel()

    def calculate_probabilities_density_matrix(self, state, qubits, nqubits):
        order = tuple(sorted(qubits))
        order += tuple(i for i in range(nqubits) if i not in qubits)
        order = order + tuple(i + nqubits for i in order)
        shape = 2 * (2 ** len(qubits), 2 ** (nqubits - len(qubits)))
        state = self.np.reshape(state, 2 * nqubits * (2,))
        state = self.np.reshape(self.np.transpose(state, order), shape)
        probs = self.np.abs(self.np.einsum("abab->a", state))
        probs = self.np.reshape(probs, len(qubits) * (2,))
        return self._order_probabilities(probs, qubits, nqubits).ravel()

    def set_seed(self, seed):
        self.np.random.seed(seed)

    def sample_shots(self, probabilities, nshots):
        return self.np.random.choice(
            range(len(probabilities)), size=nshots, p=probabilities
        )

    def aggregate_shots(self, shots):
        return self.cast(shots, dtype=shots[0].dtype)

    def samples_to_binary(self, samples, nqubits):
        ### This is faster just staying @ NumPy.
        qrange = np.arange(nqubits - 1, -1, -1, dtype=np.int32)
        samples = self.to_numpy(samples)
        return np.mod(np.right_shift(samples[:, None], qrange), 2)

    def samples_to_decimal(self, samples, nqubits):
        ### This is faster just staying @ NumPy.
        qrange = np.arange(nqubits - 1, -1, -1, dtype=np.int32)
        qrange = (2**qrange)[:, None]
        samples = np.asarray(samples.tolist())
        return np.matmul(samples, qrange)[:, 0]

    def calculate_frequencies(self, samples):
        # Samples are a list of strings so there is no advantage in using other backends
        res, counts = np.unique(samples, return_counts=True)
        res = self.to_numpy(res).tolist()
        counts = self.to_numpy(counts).tolist()
        return collections.Counter(dict(zip(res, counts)))

    def update_frequencies(self, frequencies, probabilities, nsamples):
        samples = self.sample_shots(probabilities, nsamples)
        res, counts = self.np.unique(samples, return_counts=True)
        frequencies[res] += counts
        return frequencies

    def sample_frequencies(self, probabilities, nshots):
        from qibo.config import SHOT_BATCH_SIZE

        nprobs = probabilities / self.np.sum(probabilities)
        frequencies = self.np.zeros(len(nprobs), dtype=self.np.int64)
        for _ in range(nshots // SHOT_BATCH_SIZE):
            frequencies = self.update_frequencies(frequencies, nprobs, SHOT_BATCH_SIZE)
        frequencies = self.update_frequencies(
            frequencies, nprobs, nshots % SHOT_BATCH_SIZE
        )
        return collections.Counter(
            {i: int(f) for i, f in enumerate(frequencies) if f > 0}
        )

    def apply_bitflips(self, noiseless_samples, bitflip_probabilities):
        noiseless_samples = self.cast(noiseless_samples, dtype=noiseless_samples.dtype)
        fprobs = self.cast(bitflip_probabilities, dtype="float64")
        sprobs = self.cast(np.random.random(noiseless_samples.shape), dtype="float64")
        flip_0 = self.cast(sprobs < fprobs[0], dtype=noiseless_samples.dtype)
        flip_1 = self.cast(sprobs < fprobs[1], dtype=noiseless_samples.dtype)
        noisy_samples = noiseless_samples + (1 - noiseless_samples) * flip_0
        noisy_samples = noisy_samples - noiseless_samples * flip_1
        return noisy_samples

    def partial_trace(self, state, qubits, nqubits):
        state = self.cast(state)
        state = self.np.reshape(state, nqubits * (2,))
        axes = 2 * [list(qubits)]
        rho = self.np.tensordot(state, self.np.conj(state), axes)
        shape = 2 * (2 ** (nqubits - len(qubits)),)
        return self.np.reshape(rho, shape)

    def partial_trace_density_matrix(self, state, qubits, nqubits):
        state = self.cast(state)
        state = self.np.reshape(state, 2 * nqubits * (2,))

        order = tuple(sorted(qubits))
        order += tuple(i for i in range(nqubits) if i not in qubits)
        order += tuple(i + nqubits for i in order)
        shape = 2 * (2 ** len(qubits), 2 ** (nqubits - len(qubits)))

        state = self.np.transpose(state, order)
        state = self.np.reshape(state, shape)
        return self.np.einsum("abac->bc", state)

    def calculate_norm(self, state, order=2):
        state = self.cast(state)
        return self.np.linalg.norm(state, order)

    def calculate_norm_density_matrix(self, state, order="nuc"):
        state = self.cast(state)
        return self.np.linalg.norm(state, ord=order)

    def calculate_overlap(self, state1, state2):
        return self.np.abs(self.np.sum(np.conj(self.cast(state1)) * self.cast(state2)))

    def calculate_overlap_density_matrix(self, state1, state2):
        return self.np.trace(
            self.np.matmul(self.np.conj(self.cast(state1)).T, self.cast(state2))
        )

    def calculate_eigenvalues(self, matrix, k=6):
        if self.issparse(matrix):
            log.warning(
                "Calculating sparse matrix eigenvectors because "
                "sparse modules do not provide ``eigvals`` method."
            )
            return self.calculate_eigenvectors(matrix, k=k)[0]
        return np.linalg.eigvalsh(matrix)

    def calculate_eigenvectors(self, matrix, k=6):
        if self.issparse(matrix):
            if k < matrix.shape[0]:
                from scipy.sparse.linalg import eigsh

                return eigsh(matrix, k=k, which="SA")
            else:  # pragma: no cover
                matrix = self.to_numpy(matrix)
        return np.linalg.eigh(matrix)

    def calculate_matrix_exp(self, a, matrix, eigenvectors=None, eigenvalues=None):
        if eigenvectors is None or self.issparse(matrix):
            if self.issparse(matrix):
                from scipy.sparse.linalg import expm
            else:
                from scipy.linalg import expm
            return expm(-1j * a * matrix)
        expd = self.np.diag(self.np.exp(-1j * a * eigenvalues))
        ud = self.np.transpose(np.conj(eigenvectors))
        return self.np.matmul(eigenvectors, self.np.matmul(expd, ud))

    def calculate_expectation_state(self, hamiltonian, state, normalize):
        statec = self.np.conj(state)
        hstate = hamiltonian @ state
        ev = self.np.real(self.np.sum(statec * hstate))
        if normalize:
            ev /= self.np.sum(self.np.square(self.np.abs(state)))
        return ev

    def calculate_expectation_density_matrix(self, hamiltonian, state, normalize):
        ev = self.np.real(self.np.trace(self.cast(hamiltonian @ state)))
        if normalize:
            norm = self.np.real(self.np.trace(state))
            ev /= norm
        return ev

    # TODO: remove this method
    def calculate_hamiltonian_matrix_product(self, matrix1, matrix2):
        return matrix1 @ matrix2

    # TODO: remove this method
    def calculate_hamiltonian_state_product(self, matrix, state):
        if len(tuple(state.shape)) > 2:
            raise_error(
                ValueError,
                f"Cannot multiply Hamiltonian with rank-{len(tuple(state.shape))} tensor.",
            )
        return matrix @ state

    def assert_allclose(self, value, target, rtol=1e-7, atol=0.0):
        if isinstance(value, CircuitResult) or isinstance(value, QuantumState):
            value = value.state()
        if isinstance(target, CircuitResult) or isinstance(target, QuantumState):
            target = target.state()
        value = self.to_numpy(value)
        target = self.to_numpy(target)
        np.testing.assert_allclose(value, target, rtol=rtol, atol=atol)

    def test_regressions(self, name):
        if name == "test_measurementresult_apply_bitflips":
            return [
                [0, 0, 0, 0, 2, 3, 0, 0, 0, 0],
                [0, 0, 0, 0, 2, 3, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
                [0, 0, 0, 0, 2, 0, 0, 0, 0, 0],
            ]
        elif name == "test_probabilistic_measurement":
            return {0: 249, 1: 231, 2: 253, 3: 267}
        elif name == "test_unbalanced_probabilistic_measurement":
            return {0: 171, 1: 148, 2: 161, 3: 520}
        elif name == "test_post_measurement_bitflips_on_circuit":
            return [
                {5: 30},
                {5: 18, 4: 5, 7: 4, 1: 2, 6: 1},
                {4: 8, 2: 6, 5: 5, 1: 3, 3: 3, 6: 2, 7: 2, 0: 1},
            ]



"""PyTorch backend."""

import numpy as np

from qibo import __version__
from qibo.backends.npmatrices import NumpyMatrices
from qibo.backends.numpy import NumpyBackend


class TorchMatrices(NumpyMatrices):
    """Matrix representation of every gate as a torch Tensor."""

    def __init__(self, dtype):
        import torch  # pylint: disable=import-outside-toplevel

        super().__init__(dtype)
        self.torch = torch
        self.dtype = dtype

    def _cast(self, x, dtype):
        return self.torch.as_tensor(x, dtype=dtype)

    def Unitary(self, u):
        return self._cast(u, dtype=self.dtype)


class PyTorchBackend(NumpyBackend):
    def __init__(self):
        super().__init__()
        import torch  # pylint: disable=import-outside-toplevel

        self.np = torch

        self.name = "pytorch"
        self.versions = {
            "qibo": __version__,
            "numpy": np.__version__,
            "torch": self.np.__version__,
        }

        self.dtype = self._torch_dtype(self.dtype)
        self.matrices = TorchMatrices(self.dtype)
        self.device = self.np.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.nthreads = 0
        self.tensor_types = (self.np.Tensor, np.ndarray)

        # These functions in Torch works in a different way than numpy or have different names
        self.np.transpose = self.np.permute
        self.np.expand_dims = self.np.unsqueeze
        self.np.mod = self.np.remainder
        self.np.right_shift = self.np.bitwise_right_shift

    def _torch_dtype(self, dtype):
        if dtype == "float":
            dtype += "32"
        return getattr(self.np, dtype)

    def set_device(self, device):  # pragma: no cover
        self.device = device

    def cast(
        self,
        x,
        dtype=None,
        copy: bool = False,
    ):
        """Casts input as a Torch tensor of the specified dtype.

        This method supports casting of single tensors or lists of tensors
        as for the :class:`qibo.backends.PyTorchBackend`.

        Args:
            x (Union[torch.Tensor, list[torch.Tensor], np.ndarray, list[np.ndarray], int, float, complex]):
                Input to be casted.
            dtype (Union[str, torch.dtype, np.dtype, type]): Target data type.
                If ``None``, the default dtype of the backend is used.
                Defaults to ``None``.
            copy (bool, optional): If ``True``, the input tensor is copied before casting.
                Defaults to ``False``.
        """
        if dtype is None:
            dtype = self.dtype
        elif isinstance(dtype, type):
            dtype = self._torch_dtype(dtype.__name__)
        elif not isinstance(dtype, self.np.dtype):
            dtype = self._torch_dtype(str(dtype))

        if isinstance(x, self.np.Tensor):
            x = x.to(dtype)
        elif isinstance(x, list) and all(isinstance(row, self.np.Tensor) for row in x):
            x = self.np.stack(x)
        else:
            x = self.np.tensor(x, dtype=dtype)

        if copy:
            return x.clone()

        return x

    def issparse(self, x):
        if isinstance(x, self.np.Tensor):
            return x.is_sparse

        return super().issparse(x)

    def to_numpy(self, x):
        if isinstance(x, list):
            return np.asarray([self.to_numpy(i) for i in x])

        if isinstance(x, self.np.Tensor):
            return x.numpy(force=True)

        return x

    def _order_probabilities(self, probs, qubits, nqubits):
        """Arrange probabilities according to the given ``qubits`` ordering."""
        if probs.dim() == 0:  # pragma: no cover
            return probs
        unmeasured, reduced = [], {}
        for i in range(nqubits):
            if i in qubits:
                reduced[i] = i - len(unmeasured)
            else:
                unmeasured.append(i)
        return self.np.transpose(probs, [reduced.get(i) for i in qubits])

    def calculate_probabilities(self, state, qubits, nqubits):
        rtype = self.np.real(state).dtype
        unmeasured_qubits = tuple(i for i in range(nqubits) if i not in qubits)
        state = self.np.reshape(self.np.abs(state) ** 2, nqubits * (2,))
        if len(unmeasured_qubits) == 0:
            probs = self.cast(state, dtype=rtype)
        else:
            probs = self.np.sum(self.cast(state, dtype=rtype), axis=unmeasured_qubits)
        return self._order_probabilities(probs, qubits, nqubits).ravel()

    def set_seed(self, seed):
        self.np.manual_seed(seed)
        np.random.seed(seed)

    def sample_shots(self, probabilities, nshots):
        return self.np.multinomial(
            self.cast(probabilities, dtype="float"), nshots, replacement=True
        )

    def calculate_eigenvalues(self, matrix, k=6):
        return self.np.linalg.eigvalsh(matrix)  # pylint: disable=not-callable

    def calculate_eigenvectors(self, matrix, k=6):
        return self.np.linalg.eigh(matrix)  # pylint: disable=not-callable

    def calculate_matrix_exp(self, a, matrix, eigenvectors=None, eigenvalues=None):
        if eigenvectors is None or self.issparse(matrix):
            return self.np.linalg.matrix_exp(  # pylint: disable=not-callable
                -1j * a * matrix
            )
        expd = self.np.diag(self.np.exp(-1j * a * eigenvalues))
        ud = self.np.conj(eigenvectors).T
        return self.np.matmul(eigenvectors, self.np.matmul(expd, ud))

    def test_regressions(self, name):
        if name == "test_measurementresult_apply_bitflips":
            return [
                [0, 0, 0, 0, 2, 3, 0, 0, 0, 0],
                [0, 0, 0, 0, 2, 3, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
                [0, 0, 0, 0, 2, 0, 0, 0, 0, 0],
            ]

        if name == "test_probabilistic_measurement":
            if self.device == "cuda":  # pragma: no cover
                return {0: 273, 1: 233, 2: 242, 3: 252}
            return {1: 270, 2: 248, 3: 244, 0: 238}

        if name == "test_unbalanced_probabilistic_measurement":
            if self.device == "cuda":  # pragma: no cover
                return {0: 196, 1: 153, 2: 156, 3: 495}
            return {3: 492, 2: 176, 0: 168, 1: 164}

        if name == "test_post_measurement_bitflips_on_circuit":
            return [
                {5: 30},
                {5: 17, 4: 5, 7: 4, 1: 2, 6: 2},
                {4: 9, 2: 5, 5: 5, 3: 4, 6: 4, 0: 1, 1: 1, 7: 1},
            ]



"""Module defining Error channels and NoiseModel class(es)."""

import collections
from itertools import combinations
from math import log2
from typing import Optional, Union

from qibo import gates
from qibo.config import raise_error


class KrausError:
    """Quantum error associated with the :class:`qibo.gates.KrausChannel`.

    Args:
        ops (list): List of Kraus operators of type ``np.ndarray``
            or ``tf.Tensor`` and of the same shape.

    """

    def __init__(self, ops):
        shape = ops[0].shape
        if any(o.shape != shape for o in ops):
            raise_error(
                ValueError,
                "Kraus operators of different shapes."
                "Use qibo.noise.CustomError instead.",
            )

        self.rank = shape[0]
        self.options = ops

    def channel(self, qubits, options):
        """Returns the quantum channel associated to the quantum error."""
        return [
            gates.KrausChannel(q, options)
            for q in combinations(qubits, int(log2(self.rank)))
        ]


class UnitaryError:
    """Quantum error associated with the :class:`qibo.gates.UnitaryChannel`.

    Args:
        probabilities (list): List of floats that correspond to the probability
            that each unitary Uk is applied.
        unitaries (list): List of unitary matrices as ``np.ndarray``/``tf.Tensor``
            of the same shape. Must have the same length as the given
            probabilities ``p``.

    """

    def __init__(self, probabilities, unitaries):
        shape = unitaries[0].shape
        if any(o.shape != shape for o in unitaries):
            raise_error(
                ValueError,
                "Unitary matrices have different shapes."
                "Use qibo.noise.CustomError instead.",
            )
        self.rank = shape[0]
        self.options = list(zip(probabilities, unitaries))

    def channel(self, qubits, options):
        """Returns the quantum channel associated to the quantum error."""
        return [
            gates.UnitaryChannel(q, options)
            for q in combinations(qubits, int(log2(self.rank)))
        ]


class PauliError:
    """Quantum error associated with the :class:`qibo.gates.PauliNoiseChannel`.

    Args:
        operators (list): see :class:`qibo.gates.PauliNoiseChannel`
    """

    def __init__(self, operators):
        self.options = operators

    def channel(self, qubits, options):
        """Returns the quantum channel associated to the quantum error."""
        return [gates.PauliNoiseChannel(q, options) for q in qubits]


class DepolarizingError:
    """Quantum error associated with the :class:`qibo.gates.DepolarizingChannel`.

    Args:
        options (float): see :class:`qibo.gates.DepolarizingChannel`
    """

    def __init__(self, lam):
        self.options = lam
        self.channel = gates.DepolarizingChannel


class ThermalRelaxationError:
    """Quantum error associated with the :class:`qibo.gates.ThermalRelaxationChannel`.

    Args:
        options (tuple): see :class:`qibo.gates.ThermalRelaxationChannel`
    """

    def __init__(self, t1, t2, time, excited_population=0):
        self.options = [t1, t2, time, excited_population]
        self.channel = gates.ThermalRelaxationChannel


class AmplitudeDampingError:
    """Quantum error associated with the :class:`qibo.gates.AmplitudeDampingChannel`.

    Args:
        options (float): see :class:`qibo.gates.AmplitudeDampingChannel`
    """

    def __init__(self, gamma):
        self.options = gamma
        self.channel = gates.AmplitudeDampingChannel


class PhaseDampingError:
    """Quantum error associated with the :class:`qibo.gates.PhaseDampingChannel`.

    Args:
        options (float): see :class:`qibo.gates.PhaseDampingChannel`
    """

    def __init__(self, gamma):
        self.options = gamma
        self.channel = gates.PhaseDampingChannel


class ReadoutError:
    """Quantum error associated with :class:'qibo.gates;ReadoutErrorChannel'.

    Args:
        options (array): see :class:'qibo.gates.ReadoutErrorChannel'
    """

    def __init__(self, probabilities):
        self.options = probabilities
        self.channel = gates.ReadoutErrorChannel


class ResetError:
    """Quantum error associated with the `qibo.gates.ResetChannel`.

    Args:
        options (tuple): see :class:`qibo.gates.ResetChannel`
    """

    def __init__(self, p0, p1):
        self.options = [p0, p1]
        self.channel = gates.ResetChannel


class CustomError:
    """Quantum error associated with the :class:`qibo.gates.Channel`

    Args:
        channel (:class:`qibo.gates.Channel`): any channel

    Example:

    .. testcode::

        import numpy as np
        from qibo.gates import KrausChannel
        from qibo.noise import CustomError

        # define |0><0|
        a1 = np.array([[1, 0], [0, 0]])
        # define |0><1|
        a2 = np.array([[0, 1], [0, 0]])

        # Create an Error associated with Kraus Channel
        # rho -> |0><0| rho |0><0| + |0><1| rho |0><1|
        error = CustomError(gates.KrausChannel((0,), [a1, a2]))

    """

    def __init__(self, channel):
        self.channel = channel


class NoiseModel:
    """Class for the implementation of a custom noise model.

    Example:

    .. testcode::

        from qibo import Circuit, gates
        from qibo.noise import NoiseModel, PauliError

        # Build specific noise model with 2 quantum errors:
        # - Pauli error on H only for qubit 1.
        # - Pauli error on CNOT for all the qubits.

        noise_model = NoiseModel()
        noise_model.add(PauliError([("X", 0.5)]), gates.H, 1)
        noise_model.add(PauliError([("Y", 0.5)]), gates.CNOT)

        # Generate noiseless circuit.
        circuit = Circuit(2)
        circuit.add([gates.H(0), gates.H(1), gates.CNOT(0, 1)])

        # Apply noise to the circuit according to the noise model.
        noisy_circuit = noise_model.apply(circuit)

    """

    def __init__(self):
        self.errors = collections.defaultdict(list)

    def add(
        self,
        error,
        gate: Optional[gates.Gate] = None,
        qubits: Optional[Union[int, tuple]] = None,
        conditions=None,
    ):
        """Add a quantum error for a specific gate and qubit to the noise model.

        Args:
            error: quantum error to associate with the gate. Possible choices
                are :class:`qibo.noise.PauliError`,
                :class:`qibo.noise.DepolarizingError`,
                :class:`qibo.noise.ThermalRelaxationError`,
                :class:`qibo.noise.AmplitudeDampingError`,
                :class:`qibo.noise.PhaseDampingError`,
                :class:`qibo.noise.ReadoutError`,
                :class:`qibo.noise.ResetError`,
                :class:`qibo.noise.UnitaryError`,
                :class:`qibo.noise.KrausError`, and
                :class:`qibo.noise.CustomError`.
            gate (:class:`qibo.gates.Gate`, optional): gate after which the noise will be added.
                If ``None``, the noise will be added after each gate except
                :class:`qibo.gates.Channel` and :class:`qibo.gates.M`.
            qubits (int or tuple, optional): qubits where the noise will be applied. If ``None``,
                the noise will be added after every instance of the gate.
                Defaults to ``None``.
            condition (callable, optional): function that takes :class:`qibo.gates.Gate`
                object as an input and returns ``True`` if noise should be added to it.

        Example:

        .. testcode::

            import numpy as np
            from qibo import Circuit, gates
            from qibo.noise import NoiseModel, PauliError

            # Check if a gate is RX(pi/2).
            def is_sqrt_x(gate):
                return np.pi/2 in gate.parameters

            # Build a noise model with a Pauli error on RX(pi/2) gates.
            error = PauliError(list(zip(["X", "Y", "Z"], [0.01, 0.5, 0.1])))
            noise = NoiseModel()
            noise.add(PauliError([("X", 0.5)]), gates.RX, conditions=is_sqrt_x)

            # Generate a noiseless circuit.
            circuit = Circuit(1)
            circuit.add(gates.RX(0, np.pi / 2))
            circuit.add(gates.RX(0, 3 * np.pi / 2))
            circuit.add(gates.X(0))

            # Apply noise to the circuit.
            noisy_circuit = noise.apply(circuit)

        """

        if isinstance(qubits, int):
            qubits = (qubits,)

        if (
            conditions is not None
            and not callable(conditions)
            and not isinstance(conditions, list)
        ):
            raise_error(
                TypeError,
                "`conditions` should be  either a callable or a list of callables. "
                + f"Got {type(conditions)} instead.",
            )

        if isinstance(conditions, list) and not all(
            callable(condition) for condition in conditions
        ):
            raise_error(
                TypeError,
                "A element of `conditions` list is not a callable.",
            )

        if callable(conditions):
            conditions = [conditions]

        self.errors[gate].append((conditions, error, qubits))

    def apply(self, circuit):
        """Generate a noisy quantum circuit according to the noise model built.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): quantum circuit

        Returns:
            :class:`qibo.models.circuit.Circuit`: initial circuit with noise gates
                added according to the noise model.
        """

        noisy_circuit = circuit.__class__(**circuit.init_kwargs)
        for gate in circuit.queue:
            errors_list = (
                self.errors[gate.__class__]
                if isinstance(gate, (gates.Channel, gates.M))
                else self.errors[gate.__class__] + self.errors[None]
            )

            if all(not isinstance(error, ReadoutError) for _, error, _ in errors_list):
                noisy_circuit.add(gate)

            for conditions, error, qubits in errors_list:
                if conditions is None or all(
                    condition(gate) for condition in conditions
                ):
                    qubits = (
                        gate.qubits
                        if qubits is None
                        else tuple(set(gate.qubits) & set(qubits))
                    )

                    if len(qubits) == 0:
                        continue

                    if isinstance(error, CustomError) and qubits:
                        noisy_circuit.add(error.channel)
                    elif (
                        isinstance(
                            error,
                            (
                                ThermalRelaxationError,
                                AmplitudeDampingError,
                                PhaseDampingError,
                                ResetError,
                            ),
                        )
                        and qubits
                    ):
                        for qubit in qubits:
                            noisy_circuit.add(error.channel(qubit, error.options))
                    elif isinstance(error, ReadoutError) and qubits:
                        noisy_circuit.add(error.channel(qubits, error.options))
                        noisy_circuit.add(gate)
                    else:
                        noisy_circuit.add(error.channel(qubits, error.options))

            if gate.name == "measure":
                readout_error_qubits = [
                    qubits
                    for _, error, qubits in errors_list
                    if isinstance(error, ReadoutError)
                ]
                if (
                    gate.qubits not in readout_error_qubits
                    and gate.register_name
                    not in noisy_circuit.measurement_tuples.keys()
                ):
                    noisy_circuit.add(gate)

        return noisy_circuit


class _Conditions:
    def __init__(self, qubits=None):
        self.qubits = qubits

    def condition_qubits(self, gate):
        return gate.qubits == self.qubits

    def condition_gate_single(self, gate):
        """Condition that had to be matched to apply noise channel to single-qubit ``gate``."""
        return len(gate.qubits) == 1

    def condition_gate_two(self, gate):
        """Condition that had to be matched to apply noise channel to two-qubit ``gate``."""
        return len(gate.qubits) == 2


class IBMQNoiseModel(NoiseModel):
    """Class for the implementation of a IBMQ noise model.

    This noise model applies a :class:`qibo.gates.DepolarizingChannel` followed by a
    :class:`qibo.gates.ThermalRelaxationChannel` after each one- or two-qubit gate in the circuit.
    It also applies single-qubit :class:`qibo.gates.ReadoutErrorChannel`
    *before* every measurement gate.


    Example:

    .. testcode::

        from qibo import Circuit, gates
        from qibo.models.encodings import phase_encoder
        from qibo.noise import DepolarizingError, ThermalRelaxationError, ReadoutError
        from qibo.noise import IBMQNoiseModel, NoiseModel

        nqubits = 4

        # creating circuit
        phases = list(range(nqubits))
        circuit = phase_encoder(phases, rotation="RY")
        circuit.add(gates.CNOT(qubit, qubit + 1) for qubit in range(nqubits - 1))
        circuit.add(gates.M(qubit) for qubit in range(1, nqubits - 1))

        # creating noise model from dictionary
        parameters = {
            "depolarizing_one_qubit" : {"0": 0.1, "2": 0.04, "3": 0.15},
            "depolarizing_two_qubit": {"0-1": 0.2},
            "t1" : {"0": 0.1, "1": 0.2, "3": 0.01},
            "t2" : {"0": 0.01, "1": 0.02, "3": 0.0001},
            "gate_times" : (0.1, 0.2),
            "excited_population" : 0.1,
            "readout_one_qubit" : {"0": (0.1, 0.1), "1": 0.1, "3": [0.1, 0.1]},
            }

        noise_model = IBMQNoiseModel()
        noise_model.from_dict(parameters)
        noisy_circuit = noise_model.apply(circuit)
    """

    def from_dict(self, parameters: dict):
        """Method used to pass noise ``parameters`` as inside dictionary.

        Args:
            parameters (dict): Contains parameters necessary to initialise
                :class:`qibo.noise.DepolarizingError`, :class:`qibo.noise.ThermalRelaxationError`,
                and :class:`qibo.noise.ReadoutError`.

                The keys and values of the dictionary parameters are defined below:

                - ``"depolarizing_one_qubit"`` (*int* or *float* or *dict*):  If ``int`` or
                    ``float``, all qubits share the same single-qubit depolarizing parameter.
                    If ``dict``, expects qubit indexes as keys and their respective
                    depolarizing parameter as values.
                    See :class:`qibo.gates.channels.DepolarizingChannel`
                    for a detailed definition of depolarizing parameter.
                - ``"depolarizing_two_qubit"`` (*int* or *float* or *dict*):  If ``int`` or
                    ``float``, all two-qubit gates share the same two-qubit depolarizing
                    parameter regardless in which pair of qubits the two-qubit gate is acting on.
                    If ``dict``, expects pair qubit indexes as keys separated by a hiphen
                    (e.g. "0-1" for gate that has "0" as control and "1" as target)
                    and their respective depolarizing parameter as values.
                    See :class:`qibo.gates.channels.DepolarizingChannel`
                    for a detailed definition of depolarizing parameter.
                - ``"t1"`` (*int* or *float* or *dict*): If ``int`` or ``float``, all qubits
                    share the same ``t1``. If ``dict``, expects qubit indexes as keys and its
                    respective ``t1`` as values.
                    See :class:`qibo.gates.channels.ThermalRelaxationChannel`
                    for a detailed definition of ``t1``.
                    Note that ``t1`` and ``t2`` must be passed with the same type.
                - ``"t2"`` (*int* or *float* or *dict*): If ``int`` or ``float``, all qubits share
                    the same ``t2``. If ``dict``, expects qubit indexes as keys and its
                    respective ``t2`` as values.
                    See :class:`qibo.gates.channels.ThermalRelaxationChannel`
                    for a detailed definition of ``t2``.
                    Note that ``t2`` and ``t1`` must be passed with the same type.
                - ``"gate_times"`` (*tuple* or *list*): pair of gate times representing
                    gate times for :class:`ThermalRelaxationError` following, respectively,
                    one- and two-qubit gates.
                - ``"excited_population"`` (*int* or *float*): See
                    :class:`ThermalRelaxationChannel`.
                - ``"readout_one_qubit"`` (*int* or *float* or *dict*): If ``int`` or ``float``,
                    :math:`p(0|1) = p(1|0)`, and all qubits share the same readout error
                    probabilities. If ``dict``, expects qubit indexes as keys and
                    values as ``tuple`` (or ``list``) in the format :math:`(p(0|1),\\,p(1|0))`.
                    If values are ``tuple`` or ``list`` of length 1 or ``float`` or ``int``,
                    then it is assumed that :math:`p(0|1) = p(1|0)`.
        """
        t_1 = parameters["t1"]
        t_2 = parameters["t2"]
        gate_time_1, gate_time_2 = parameters["gate_times"]
        excited_population = parameters["excited_population"]
        depolarizing_one_qubit = parameters["depolarizing_one_qubit"]
        depolarizing_two_qubit = parameters["depolarizing_two_qubit"]
        readout_one_qubit = parameters["readout_one_qubit"]

        if isinstance(depolarizing_one_qubit, (float, int)):
            self.add(
                DepolarizingError(depolarizing_one_qubit),
                conditions=_Conditions().condition_gate_single,
            )

        if isinstance(depolarizing_one_qubit, dict):
            for qubit_key, lamb in depolarizing_one_qubit.items():
                self.add(
                    DepolarizingError(lamb),
                    qubits=int(qubit_key),
                    conditions=_Conditions().condition_gate_single,
                )

        if isinstance(depolarizing_two_qubit, (float, int)):
            self.add(
                DepolarizingError(depolarizing_two_qubit),
                conditions=_Conditions().condition_gate_two,
            )

        if isinstance(depolarizing_two_qubit, dict):
            for key, lamb in depolarizing_two_qubit.items():
                qubits = key.replace(" ", "").split("-")
                qubits = tuple(map(int, qubits))
                self.add(
                    DepolarizingError(lamb),
                    qubits=qubits,
                    conditions=[
                        _Conditions().condition_gate_two,
                        _Conditions(qubits).condition_qubits,
                    ],
                )

        if isinstance(t_1, (float, int)) and isinstance(t_2, (float, int)):
            self.add(
                ThermalRelaxationError(t_1, t_2, gate_time_1, excited_population),
                conditions=_Conditions().condition_gate_single,
            )
            self.add(
                ThermalRelaxationError(t_1, t_2, gate_time_2, excited_population),
                conditions=_Conditions().condition_gate_two,
            )

        if isinstance(t_1, dict) and isinstance(t_2, dict):
            for qubit_key in t_1.keys():
                self.add(
                    ThermalRelaxationError(
                        t_1[qubit_key], t_2[qubit_key], gate_time_1, excited_population
                    ),
                    qubits=int(qubit_key),
                    conditions=_Conditions().condition_gate_single,
                )
                self.add(
                    ThermalRelaxationError(
                        t_1[qubit_key], t_2[qubit_key], gate_time_2, excited_population
                    ),
                    qubits=int(qubit_key),
                    conditions=_Conditions().condition_gate_two,
                )

        if isinstance(readout_one_qubit, (int, float)):
            probabilities = [
                [1 - readout_one_qubit, readout_one_qubit],
                [readout_one_qubit, 1 - readout_one_qubit],
            ]
            self.add(ReadoutError(probabilities), gate=gates.M)

        if isinstance(readout_one_qubit, dict):
            for qubit, probs in readout_one_qubit.items():
                if isinstance(probs, (int, float)):
                    probs = (probs, probs)
                elif isinstance(probs, (tuple, list)) and len(probs) == 1:
                    probs *= 2

                probabilities = [[1 - probs[0], probs[0]], [probs[1], 1 - probs[1]]]
                self.add(
                    ReadoutError(probabilities),
                    gate=gates.M,
                    qubits=int(qubit),
                )



"""Module defining Hamiltonian classes."""

from itertools import chain
from typing import Optional

import numpy as np
import sympy

from qibo.backends import PyTorchBackend
from qibo.config import EINSUM_CHARS, log, raise_error
from qibo.hamiltonians.abstract import AbstractHamiltonian
from qibo.symbols import Z


class Hamiltonian(AbstractHamiltonian):
    """Hamiltonian based on a dense or sparse matrix representation.

    Args:
        nqubits (int): number of quantum bits.
        matrix (np.ndarray): Matrix representation of the Hamiltonian in the
            computational basis as an array of shape ``(2 ** nqubits, 2 ** nqubits)``.
            Sparse matrices based on ``scipy.sparse`` for numpy/qibojit backends
            or on ``tf.sparse`` for the tensorflow backend are also
            supported.
    """

    def __init__(self, nqubits, matrix=None, backend=None):
        from qibo.backends import _check_backend

        self.backend = _check_backend(backend)

        if not (
            isinstance(matrix, self.backend.tensor_types)
            or self.backend.issparse(matrix)
        ):
            raise_error(
                TypeError,
                f"Matrix of invalid type {type(matrix)} given during Hamiltonian initialization",
            )
        matrix = self.backend.cast(matrix)

        super().__init__()
        self.nqubits = nqubits
        self.matrix = matrix
        self._eigenvalues = None
        self._eigenvectors = None
        self._exp = {"a": None, "result": None}

    @property
    def matrix(self):
        """Returns the full matrix representation.

        Can be a dense ``(2 ** nqubits, 2 ** nqubits)`` array or a sparse
        matrix, depending on how the Hamiltonian was created.
        """
        return self._matrix

    @matrix.setter
    def matrix(self, mat):
        shape = tuple(mat.shape)
        if shape != 2 * (2**self.nqubits,):
            raise_error(
                ValueError,
                f"The Hamiltonian is defined for {self.nqubits} qubits "
                + f"while the given matrix has shape {shape}.",
            )
        self._matrix = mat

    @classmethod
    def from_symbolic(cls, symbolic_hamiltonian, symbol_map, backend=None):
        """Creates a ``Hamiltonian`` from a symbolic Hamiltonian.

        We refer to the
        :ref:`How to define custom Hamiltonians using symbols? <symbolicham-example>`
        example for more details.

        Args:
            symbolic_hamiltonian (sympy.Expr): The full Hamiltonian written
                with symbols.
            symbol_map (dict): Dictionary that maps each symbol that appears in
                the Hamiltonian to a pair of (target, matrix).

        Returns:
            A :class:`qibo.hamiltonians.SymbolicHamiltonian` object
            that implements the Hamiltonian represented by the given symbolic
            expression.
        """
        log.warning(
            "`Hamiltonian.from_symbolic` and the use of symbol maps is "
            "deprecated. Please use `SymbolicHamiltonian` and Qibo symbols "
            "to construct Hamiltonians using symbols."
        )
        return SymbolicHamiltonian(
            symbolic_hamiltonian, symbol_map=symbol_map, backend=backend
        )

    def eigenvalues(self, k=6):
        if self._eigenvalues is None:
            self._eigenvalues = self.backend.calculate_eigenvalues(self.matrix, k)
        return self._eigenvalues

    def eigenvectors(self, k=6):
        if self._eigenvectors is None:
            self._eigenvalues, self._eigenvectors = self.backend.calculate_eigenvectors(
                self.matrix, k
            )
        return self._eigenvectors

    def exp(self, a):
        if self._exp.get("a") != a:
            self._exp["a"] = a
            self._exp["result"] = self.backend.calculate_matrix_exp(
                a, self.matrix, self._eigenvectors, self._eigenvalues
            )
        return self._exp.get("result")

    def expectation(self, state, normalize=False):
        if isinstance(state, self.backend.tensor_types):
            state = self.backend.cast(state)
            shape = tuple(state.shape)
            if len(shape) == 1:  # state vector
                return self.backend.calculate_expectation_state(self, state, normalize)

            if len(shape) == 2:  # density matrix
                return self.backend.calculate_expectation_density_matrix(
                    self, state, normalize
                )

            raise_error(
                ValueError,
                "Cannot calculate Hamiltonian expectation value "
                + f"for state of shape {shape}",
            )

        raise_error(
            TypeError,
            "Cannot calculate Hamiltonian expectation "
            + f"value for state of type {type(state)}",
        )

    def expectation_from_samples(self, freq, qubit_map=None):
        obs = self.matrix
        if np.count_nonzero(obs - np.diag(np.diagonal(obs))) != 0:
            raise_error(NotImplementedError, "Observable is not diagonal.")
        keys = list(freq.keys())
        if qubit_map is None:
            qubit_map = list(range(int(np.log2(len(obs)))))
        counts = np.array(list(freq.values())) / sum(freq.values())
        expval = 0
        size = len(qubit_map)
        for j, k in enumerate(keys):
            index = 0
            for i in qubit_map:
                index += int(k[qubit_map.index(i)]) * 2 ** (size - 1 - i)
            expval += obs[index, index] * counts[j]
        return np.real(expval)

    def eye(self, dim: Optional[int] = None):
        """Generate Identity matrix with dimension ``dim``"""
        if dim is None:
            dim = int(self.matrix.shape[0])
        return self.backend.cast(self.backend.matrices.I(dim), dtype=self.matrix.dtype)

    def energy_fluctuation(self, state):
        """
        Evaluate energy fluctuation:

        .. math::
            \\Xi_{k}(\\mu) = \\sqrt{\\langle\\mu|\\hat{H}^2|\\mu\\rangle - \\langle\\mu|\\hat{H}|\\mu\\rangle^2} \\,

        for a given state :math:`|\\mu\\rangle`.

        Args:
            state (np.ndarray): quantum state to be used to compute the energy fluctuation.

        Return:
            Energy fluctuation value (float).
        """
        state = self.backend.cast(state)
        energy = self.expectation(state)
        h = self.matrix
        h2 = Hamiltonian(nqubits=self.nqubits, matrix=h @ h, backend=self.backend)
        average_h2 = h2.expectation(state, normalize=True)
        return np.sqrt(np.abs(average_h2 - energy**2))

    def __add__(self, o):
        if isinstance(o, self.__class__):
            if self.nqubits != o.nqubits:
                raise_error(
                    RuntimeError,
                    "Only hamiltonians with the same number of qubits can be added.",
                )
            new_matrix = self.matrix + o.matrix
        elif isinstance(o, self.backend.numeric_types):
            new_matrix = self.matrix + o * self.eye()
        else:
            raise_error(
                NotImplementedError,
                f"Hamiltonian addition to {type(o)} not implemented.",
            )
        return self.__class__(self.nqubits, new_matrix, backend=self.backend)

    def __sub__(self, o):
        if isinstance(o, self.__class__):
            if self.nqubits != o.nqubits:
                raise_error(
                    RuntimeError,
                    "Only hamiltonians with the same number of qubits can be subtracted.",
                )
            new_matrix = self.matrix - o.matrix
        elif isinstance(o, self.backend.numeric_types):
            new_matrix = self.matrix - o * self.eye()
        else:
            raise_error(
                NotImplementedError,
                f"Hamiltonian subtraction to {type(o)} not implemented.",
            )
        return self.__class__(self.nqubits, new_matrix, backend=self.backend)

    def __rsub__(self, o):
        if isinstance(o, self.__class__):  # pragma: no cover
            # impractical case because it will be handled by `__sub__`
            if self.nqubits != o.nqubits:
                raise_error(
                    RuntimeError,
                    "Only hamiltonians with the same number of qubits can be added.",
                )
            new_matrix = o.matrix - self.matrix
        elif isinstance(o, self.backend.numeric_types):
            new_matrix = o * self.eye() - self.matrix
        else:
            raise_error(
                NotImplementedError,
                f"Hamiltonian subtraction to {type(o)} not implemented.",
            )
        return self.__class__(self.nqubits, new_matrix, backend=self.backend)

    def __mul__(self, o):
        if isinstance(o, self.backend.tensor_types):
            o = complex(o)
        elif not isinstance(o, self.backend.numeric_types):
            raise_error(
                NotImplementedError,
                f"Hamiltonian multiplication to {type(o)} not implemented.",
            )
        new_matrix = self.matrix * o
        r = self.__class__(self.nqubits, new_matrix, backend=self.backend)
        o = self.backend.cast(o)
        if self._eigenvalues is not None:
            if self.backend.np.real(o) >= 0:  # TODO: check for side effects K.qnp
                r._eigenvalues = o * self._eigenvalues
            elif not self.backend.issparse(self.matrix):
                axis = (0,) if isinstance(self.backend, PyTorchBackend) else 0
                r._eigenvalues = o * self.backend.np.flip(self._eigenvalues, axis)
        if self._eigenvectors is not None:
            if self.backend.np.real(o) > 0:  # TODO: see above
                r._eigenvectors = self._eigenvectors
            elif o == 0:
                r._eigenvectors = self.eye(int(self._eigenvectors.shape[0]))
        return r

    def __matmul__(self, o):
        if isinstance(o, self.__class__):
            matrix = self.backend.calculate_hamiltonian_matrix_product(
                self.matrix, o.matrix
            )
            return self.__class__(self.nqubits, matrix, backend=self.backend)

        if isinstance(o, self.backend.tensor_types):
            return self.backend.calculate_hamiltonian_state_product(self.matrix, o)

        raise_error(
            NotImplementedError,
            f"Hamiltonian matmul to {type(o)} not implemented.",
        )


class SymbolicHamiltonian(AbstractHamiltonian):
    """Hamiltonian based on a symbolic representation.

    Calculations using symbolic Hamiltonians are either done directly using
    the given ``sympy`` expression as it is (``form``) or by parsing the
    corresponding ``terms`` (which are :class:`qibo.core.terms.SymbolicTerm`
    objects). The latter approach is more computationally costly as it uses
    a ``sympy.expand`` call on the given form before parsing the terms.
    For this reason the ``terms`` are calculated only when needed, for example
    during Trotterization.
    The dense matrix of the symbolic Hamiltonian can be calculated directly
    from ``form`` without requiring ``terms`` calculation (see
    :meth:`qibo.core.hamiltonians.SymbolicHamiltonian.calculate_dense` for details).

    Args:
        form (sympy.Expr): Hamiltonian form as a ``sympy.Expr``. Ideally the
            Hamiltonian should be written using Qibo symbols.
            See :ref:`How to define custom Hamiltonians using symbols? <symbolicham-example>`
            example for more details.
        symbol_map (dict): Dictionary that maps each ``sympy.Symbol`` to a tuple
            of (target qubit, matrix representation). This feature is kept for
            compatibility with older versions where Qibo symbols were not available
            and may be deprecated in the future.
            It is not required if the Hamiltonian is constructed using Qibo symbols.
            The symbol_map can also be used to pass non-quantum operator arguments
            to the symbolic Hamiltonian, such as the parameters in the
            :meth:`qibo.hamiltonians.models.MaxCut` Hamiltonian.
    """

    def __init__(self, form=None, nqubits=None, symbol_map={}, backend=None):
        super().__init__()
        self._form = None
        self._terms = None
        self.constant = 0  # used only when we perform calculations using ``_terms``
        self._dense = None
        self.symbol_map = symbol_map
        # if a symbol in the given form is not a Qibo symbol it must be
        # included in the ``symbol_map``

        from qibo.symbols import Symbol  # pylint: disable=import-outside-toplevel

        self._qiboSymbol = Symbol  # also used in ``self._get_symbol_matrix``

        from qibo.backends import _check_backend

        self.backend = _check_backend(backend)

        if form is not None:
            self.form = form
        if nqubits is not None:
            self.nqubits = nqubits

    @property
    def dense(self):
        """Creates the equivalent :class:`qibo.hamiltonians.MatrixHamiltonian`."""
        if self._dense is None:
            log.warning(
                "Calculating the dense form of a symbolic Hamiltonian. "
                "This operation is memory inefficient."
            )
            self.dense = self.calculate_dense()
        return self._dense

    @dense.setter
    def dense(self, hamiltonian):
        assert isinstance(hamiltonian, Hamiltonian)
        self._dense = hamiltonian
        self._eigenvalues = hamiltonian._eigenvalues
        self._eigenvectors = hamiltonian._eigenvectors
        self._exp = hamiltonian._exp

    @property
    def form(self):
        return self._form

    @form.setter
    def form(self, form):
        # Check that given form is a ``sympy`` expression
        if not isinstance(form, sympy.Expr):
            raise_error(
                TypeError,
                f"Symbolic Hamiltonian should be a ``sympy`` expression but is {type(form)}.",
            )
        # Calculate number of qubits in the system described by the given
        # Hamiltonian formula
        nqubits = 0
        for symbol in form.free_symbols:
            if isinstance(symbol, self._qiboSymbol):
                q = symbol.target_qubit
            elif isinstance(symbol, sympy.Expr):
                if symbol not in self.symbol_map:
                    raise_error(ValueError, f"Symbol {symbol} is not in symbol map.")
                q, matrix = self.symbol_map.get(symbol)
                if not isinstance(matrix, self.backend.tensor_types):
                    # ignore symbols that do not correspond to quantum operators
                    # for example parameters in the MaxCut Hamiltonian
                    q = 0
            if q > nqubits:
                nqubits = q

        self._form = form
        self.nqubits = nqubits + 1

    @property
    def terms(self):
        """List of :class:`qibo.core.terms.HamiltonianTerm` objects
        of which the Hamiltonian is a sum of.
        """
        if self._terms is None:
            # Calculate terms based on ``self.form``
            from qibo.hamiltonians.terms import (  # pylint: disable=import-outside-toplevel
                SymbolicTerm,
            )

            form = sympy.expand(self.form)
            terms = []
            for f, c in form.as_coefficients_dict().items():
                term = SymbolicTerm(c, f, self.symbol_map)
                if term.target_qubits:
                    terms.append(term)
                else:
                    self.constant += term.coefficient
            self._terms = terms
        return self._terms

    @terms.setter
    def terms(self, terms):
        self._terms = terms
        self.nqubits = max(q for term in self._terms for q in term.target_qubits) + 1

    @property
    def matrix(self):
        """Returns the full ``(2 ** nqubits, 2 ** nqubits)`` matrix representation."""
        return self.dense.matrix

    def eigenvalues(self, k=6):
        return self.dense.eigenvalues(k)

    def eigenvectors(self, k=6):
        return self.dense.eigenvectors(k)

    def ground_state(self):
        return self.eigenvectors()[:, 0]

    def exp(self, a):
        return self.dense.exp(a)

    def _get_symbol_matrix(self, term):
        """Calculates numerical matrix corresponding to symbolic expression.

        This is partly equivalent to sympy's ``.subs``, which does not work
        in our case as it does not allow us to substitute ``sympy.Symbol``
        with numpy arrays and there are different complication when switching
        to ``sympy.MatrixSymbol``. Here we calculate the full numerical matrix
        given the symbolic expression using recursion.
        Helper method for ``_calculate_dense_from_form``.

        Args:
            term (sympy.Expr): Symbolic expression containing local operators.

        Returns:
            Numerical matrix corresponding to the given expression as a numpy
            array of size ``(2 ** self.nqubits, 2 ** self.nqubits).
        """
        if isinstance(term, sympy.Add):
            # symbolic op for addition
            result = sum(
                self._get_symbol_matrix(subterm) for subterm in term.as_ordered_terms()
            )

        elif isinstance(term, sympy.Mul):
            # symbolic op for multiplication
            # note that we need to use matrix multiplication even though
            # we use scalar symbols for convenience
            factors = term.as_ordered_factors()
            result = self._get_symbol_matrix(factors[0])
            for subterm in factors[1:]:
                result = result @ self._get_symbol_matrix(subterm)

        elif isinstance(term, sympy.Pow):
            # symbolic op for power
            base, exponent = term.as_base_exp()
            matrix = self._get_symbol_matrix(base)
            # multiply ``base`` matrix ``exponent`` times to itself
            result = matrix
            for _ in range(exponent - 1):
                result = result @ matrix

        elif isinstance(term, sympy.Symbol):
            # if the term is a ``Symbol`` then it corresponds to a quantum
            # operator for which we can construct the full matrix directly
            if isinstance(term, self._qiboSymbol):
                # if we have a Qibo symbol the matrix construction is
                # implemented in :meth:`qibo.core.terms.SymbolicTerm.full_matrix`.
                result = term.full_matrix(self.nqubits)
            else:
                q, matrix = self.symbol_map.get(term)
                if not isinstance(matrix, self.backend.tensor_types):
                    # symbols that do not correspond to quantum operators
                    # for example parameters in the MaxCut Hamiltonian
                    result = complex(matrix) * np.eye(2**self.nqubits)
                else:
                    # if we do not have a Qibo symbol we construct one and use
                    # :meth:`qibo.core.terms.SymbolicTerm.full_matrix`.
                    result = self._qiboSymbol(q, matrix).full_matrix(self.nqubits)

        elif term.is_number:
            # if the term is number we should return in the form of identity
            # matrix because in expressions like `1 + Z`, `1` is not correspond
            # to the float 1 but the identity operator (matrix)
            result = complex(term) * np.eye(2**self.nqubits)

        else:
            raise_error(
                TypeError,
                f"Cannot calculate matrix for symbolic term of type {type(term)}.",
            )

        return result

    def _calculate_dense_from_form(self):
        """Calculates equivalent :class:`qibo.core.hamiltonians.Hamiltonian` using symbolic form.
        Useful when the term representation is not available.
        """
        matrix = self._get_symbol_matrix(self.form)
        return Hamiltonian(self.nqubits, matrix, backend=self.backend)

    def _calculate_dense_from_terms(self):
        """Calculates equivalent :class:`qibo.core.hamiltonians.Hamiltonian`
        using the term representation.
        """
        if 2 * self.nqubits > len(EINSUM_CHARS):  # pragma: no cover
            # case not tested because it only happens in large examples
            raise_error(NotImplementedError, "Not enough einsum characters.")

        matrix = 0
        chars = EINSUM_CHARS[: 2 * self.nqubits]
        for term in self.terms:
            ntargets = len(term.target_qubits)
            tmat = np.reshape(term.matrix, 2 * ntargets * (2,))
            n = self.nqubits - ntargets
            emat = np.reshape(np.eye(2**n, dtype=tmat.dtype), 2 * n * (2,))
            gen = lambda x: (chars[i + x] for i in term.target_qubits)
            tc = "".join(chain(gen(0), gen(self.nqubits)))
            ec = "".join(c for c in chars if c not in tc)
            matrix += np.einsum(f"{tc},{ec}->{chars}", tmat, emat)
        matrix = np.reshape(matrix, 2 * (2**self.nqubits,))
        return Hamiltonian(self.nqubits, matrix, backend=self.backend) + self.constant

    def calculate_dense(self):
        if self._terms is None:
            # calculate dense matrix directly using the form to avoid the
            # costly ``sympy.expand`` call
            return self._calculate_dense_from_form()
        return self._calculate_dense_from_terms()

    def expectation(self, state, normalize=False):
        return Hamiltonian.expectation(self, state, normalize)

    def expectation_from_samples(self, freq, qubit_map=None):
        terms = self.terms
        for term in terms:
            # pylint: disable=E1101
            for factor in term.factors:
                if not isinstance(factor, Z):
                    raise_error(
                        NotImplementedError, "Observable is not a Z Pauli string."
                    )
            if len(term.factors) != len(set(term.factors)):
                raise_error(NotImplementedError, "Z^k is not implemented since Z^2=I.")
        keys = list(freq.keys())
        counts = np.array(list(freq.values())) / sum(freq.values())
        coeff = list(self.form.as_coefficients_dict().values())
        qubits = []
        for term in terms:
            qubits_term = []
            for k in term.target_qubits:
                qubits_term.append(k)
            qubits.append(qubits_term)
        if qubit_map is None:
            qubit_map = list(range(len(keys[0])))
        expval = 0
        for j, q in enumerate(qubits):
            subk = []
            expval_q = 0
            for i, k in enumerate(keys):
                subk = [int(k[qubit_map.index(s)]) for s in q]
                expval_k = 1
                if subk.count(1) % 2 == 1:
                    expval_k = -1
                expval_q += expval_k * counts[i]
            expval += expval_q * float(coeff[j])
        return expval

    def __add__(self, o):
        if isinstance(o, self.__class__):
            if self.nqubits != o.nqubits:
                raise_error(
                    RuntimeError,
                    "Only hamiltonians with the same number of qubits can be added.",
                )
            new_ham = self.__class__(
                symbol_map=dict(self.symbol_map), backend=self.backend
            )
            if self._form is not None and o._form is not None:
                new_ham.form = self.form + o.form
                new_ham.symbol_map.update(o.symbol_map)
            if self._terms is not None and o._terms is not None:
                new_ham.terms = self.terms + o.terms
                new_ham.constant = self.constant + o.constant
            if self._dense is not None and o._dense is not None:
                new_ham.dense = self.dense + o.dense

        elif isinstance(o, self.backend.numeric_types):
            new_ham = self.__class__(
                symbol_map=dict(self.symbol_map), backend=self.backend
            )
            if self._form is not None:
                new_ham.form = self.form + o
            if self._terms is not None:
                new_ham.terms = self.terms
                new_ham.constant = self.constant + o
            if self._dense is not None:
                new_ham.dense = self.dense + o

        else:
            raise_error(
                NotImplementedError,
                f"SymbolicHamiltonian addition to {type(o)} not implemented.",
            )
        return new_ham

    def __sub__(self, o):
        if isinstance(o, self.__class__):
            if self.nqubits != o.nqubits:
                raise_error(
                    RuntimeError,
                    "Only hamiltonians with the same number of qubits can be subtracted.",
                )
            new_ham = self.__class__(
                symbol_map=dict(self.symbol_map), backend=self.backend
            )
            if self._form is not None and o._form is not None:
                new_ham.form = self.form - o.form
                new_ham.symbol_map.update(o.symbol_map)
            if self._terms is not None and o._terms is not None:
                new_ham.terms = self.terms + [-1 * x for x in o.terms]
                new_ham.constant = self.constant - o.constant
            if self._dense is not None and o._dense is not None:
                new_ham.dense = self.dense - o.dense

        elif isinstance(o, self.backend.numeric_types):
            new_ham = self.__class__(
                symbol_map=dict(self.symbol_map), backend=self.backend
            )
            if self._form is not None:
                new_ham.form = self.form - o
            if self._terms is not None:
                new_ham.terms = self.terms
                new_ham.constant = self.constant - o
            if self._dense is not None:
                new_ham.dense = self.dense - o

        else:
            raise_error(
                NotImplementedError,
                f"Hamiltonian subtraction to {type(o)} " "not implemented.",
            )
        return new_ham

    def __rsub__(self, o):
        if isinstance(o, self.backend.numeric_types):
            new_ham = self.__class__(
                symbol_map=dict(self.symbol_map), backend=self.backend
            )
            if self._form is not None:
                new_ham.form = o - self.form
            if self._terms is not None:
                new_ham.terms = [-1 * x for x in self.terms]
                new_ham.constant = o - self.constant
            if self._dense is not None:
                new_ham.dense = o - self.dense
        else:
            raise_error(
                NotImplementedError,
                f"Hamiltonian subtraction to {type(o)} not implemented.",
            )
        return new_ham

    def __mul__(self, o):
        if not isinstance(o, (self.backend.numeric_types, self.backend.tensor_types)):
            raise_error(
                NotImplementedError,
                f"Hamiltonian multiplication to {type(o)} not implemented.",
            )
        o = complex(o)
        new_ham = self.__class__(symbol_map=dict(self.symbol_map), backend=self.backend)
        if self._form is not None:
            new_ham.form = o * self.form
        if self._terms is not None:
            new_ham.terms = [o * x for x in self.terms]
            new_ham.constant = self.constant * o
        if self._dense is not None:
            new_ham.dense = o * self._dense
        return new_ham

    def apply_gates(self, state, density_matrix=False):
        """Applies gates corresponding to the Hamiltonian terms to a given state.
        Helper method for ``__matmul__``.
        """
        total = 0
        for term in self.terms:
            total += term(
                self.backend,
                self.backend.cast(state, copy=True),
                self.nqubits,
                density_matrix=density_matrix,
            )
        if self.constant:  # pragma: no cover
            total += self.constant * state
        return total

    def __matmul__(self, o):
        """Matrix multiplication with other Hamiltonians or state vectors."""
        if isinstance(o, self.__class__):
            if self._form is None or o._form is None:
                raise_error(
                    NotImplementedError,
                    "Multiplication of symbolic Hamiltonians "
                    "without symbolic form is not implemented.",
                )
            new_form = self.form * o.form
            new_symbol_map = dict(self.symbol_map)
            new_symbol_map.update(o.symbol_map)
            new_ham = self.__class__(
                new_form, symbol_map=new_symbol_map, backend=self.backend
            )
            if self._dense is not None and o._dense is not None:
                new_ham.dense = self.dense @ o.dense
            return new_ham

        if isinstance(o, self.backend.tensor_types):
            rank = len(tuple(o.shape))
            if rank not in (1, 2):
                raise_error(
                    NotImplementedError,
                    f"Cannot multiply Hamiltonian with rank-{rank} tensor.",
                )
            state_qubits = int(np.log2(int(o.shape[0])))
            if state_qubits != self.nqubits:
                raise_error(
                    ValueError,
                    f"Cannot multiply Hamiltonian on {self.nqubits} qubits to "
                    + f"state of {state_qubits} qubits.",
                )
            if rank == 1:  # state vector
                return self.apply_gates(o)

            return self.apply_gates(o, density_matrix=True)

        raise_error(
            NotImplementedError,
            f"Hamiltonian matmul to {type(o)} not implemented.",
        )

    def circuit(self, dt, accelerators=None):
        """Circuit that implements a Trotter step of this Hamiltonian.

        Args:
            dt (float): Time step used for Trotterization.
            accelerators (dict): Dictionary with accelerators for distributed circuits.
        """
        from qibo import Circuit  # pylint: disable=import-outside-toplevel
        from qibo.hamiltonians.terms import (  # pylint: disable=import-outside-toplevel
            TermGroup,
        )

        groups = TermGroup.from_terms(self.terms)
        circuit = Circuit(self.nqubits, accelerators=accelerators)
        circuit.add(
            group.term.expgate(dt / 2.0) for group in chain(groups, groups[::-1])
        )

        return circuit


class TrotterHamiltonian:
    """"""

    def __init__(self, *parts):
        raise_error(
            NotImplementedError,
            "`TrotterHamiltonian` is substituted by `SymbolicHamiltonian` "
            + "and is no longer supported. Please check the documentation "
            + "of `SymbolicHamiltonian` for more details.",
        )

    @classmethod
    def from_symbolic(cls, symbolic_hamiltonian, symbol_map):
        return cls()



import hyperopt

from qibo.backends import _check_backend
from qibo.models.dbi.double_bracket import *
from qibo.models.dbi.utils import *


def select_best_dbr_generator(
    dbi_object: DoubleBracketIteration,
    d_list: list,
    step: Optional[float] = None,
    compare_canonical: bool = True,
    scheduling: DoubleBracketScheduling = None,
    **kwargs,
):
    """Selects the best double bracket rotation generator from a list and execute the rotation.

    Args:
        dbi_object (`DoubleBracketIteration`): the target DoubleBracketIteration object.
        d_list (list): list of diagonal operators (np.array) to select from.
        step (float): fixed iteration duration.
            Defaults to ``None``, optimize with `scheduling` method and `choose_step` function.
        compare_canonical (boolean): if `True`, the diagonalization effect with operators from `d_list` is compared with the canonical bracket.
        scheduling (`DoubleBracketScheduling`): scheduling method for finding the optimal step.

    Returns:
        The updated dbi_object (`DoubleBracketIteration`), index of the optimal diagonal operator (int), respective step duration (float), and sign (int).

    Example:
        from qibo.hamiltonians import Hamiltonian
        from qibo.models.dbi.double_bracket import *
        from qibo.models.dbi.utils_dbr_strategies import select_best_dbr_generator
        from qibo.quantum_info import random_hermitian

        nqubits = 3
        NSTEPS = 3
        h0 = random_hermitian(2**nqubits)
        dbi = DoubleBracketIteration(
            Hamiltonian(nqubits, h0),
            mode=DoubleBracketGeneratorType.single_commutator,
        )
        initial_off_diagonal_norm = dbi.off_diagonal_norm
        generate_local_Z = generate_Z_operators(nqubits)
        Z_ops = list(generate_local_Z.values())
        for _ in range(NSTEPS):
            dbi, idx, step, flip_sign = select_best_dbr_generator(
                dbi, Z_ops, compare_canonical=True
                )
    """
    if scheduling is None:
        scheduling = dbi_object.scheduling

    if compare_canonical:
        norms_off_diagonal_restriction = [dbi_object.off_diagonal_norm] * (
            len(d_list) + 1
        )
        optimal_steps = np.zeros(len(d_list) + 1)
        flip_list = np.ones(len(d_list) + 1)
    else:
        norms_off_diagonal_restriction = [dbi_object.off_diagonal_norm] * (len(d_list))
        optimal_steps = np.zeros(len(d_list))
        flip_list = np.ones(len(d_list))

    for i, d in enumerate(d_list):
        # prescribed step durations
        dbi_eval = deepcopy(dbi_object)
        d = dbi_eval.backend.cast(d)
        flip_list[i] = cs_angle_sgn(dbi_eval, d)
        if flip_list[i] != 0:
            if step is None:
                step_best = dbi_eval.choose_step(
                    d=flip_list[i] * d, scheduling=scheduling, **kwargs
                )
            else:
                step_best = step
            dbi_eval(step=step_best, d=flip_list[i] * d)
            optimal_steps[i] = step_best
            norms_off_diagonal_restriction[i] = dbi_eval.off_diagonal_norm
    # canonical
    if compare_canonical is True:
        dbi_eval = deepcopy(dbi_object)
        dbi_eval.mode = DoubleBracketGeneratorType.canonical
        if step is None:
            step_best = dbi_eval.choose_step(scheduling=scheduling, **kwargs)
        else:
            step_best = step
        dbi_eval(step=step_best)
        optimal_steps[-1] = step_best
        norms_off_diagonal_restriction[-1] = dbi_eval.off_diagonal_norm
    # find best d
    idx_max_loss = np.argmin(norms_off_diagonal_restriction)
    flip = flip_list[idx_max_loss]
    step_optimal = optimal_steps[idx_max_loss]
    dbi_eval = deepcopy(dbi_object)
    if idx_max_loss == len(d_list) and compare_canonical is True:
        # canonical
        dbi_eval(step=step_optimal, mode=DoubleBracketGeneratorType.canonical)

    else:
        d_optimal = flip * d_list[idx_max_loss]
        dbi_eval(step=step_optimal, d=d_optimal)
    return dbi_eval, idx_max_loss, step_optimal, flip


def gradient_numerical(
    dbi_object: DoubleBracketIteration,
    d_params: list,
    parameterization: ParameterizationTypes,
    s: float = 1e-2,
    delta: float = 1e-3,
    backend=None,
    **kwargs,
):
    r"""
    Gradient of the DBI with respect to the parametrization of D. A simple finite difference is used to calculate the gradient.

    Args:
        dbi_object (DoubleBracketIteration): DoubleBracketIteration object.
        d_params (np.array): Parameters for the ansatz (note that the dimension must be 2**nqubits for full ansazt and nqubits for Pauli ansatz).
        s (float): A short flow duration for finding the numerical gradient.
        delta (float): Step size for numerical gradient.
    Returns:
        grad (np.array): Gradient of the D operator.
    """
    backend = _check_backend(backend)
    nqubits = dbi_object.nqubits
    grad = np.zeros(len(d_params))
    d = params_to_diagonal_operator(
        d_params, nqubits, parameterization=parameterization, **kwargs
    )
    for i in range(len(d_params)):
        params_new = backend.cast(d_params, copy=True)
        params_new[i] += delta
        d_new = params_to_diagonal_operator(
            params_new, nqubits, parameterization=parameterization, **kwargs
        )
        # find the increment of a very small step
        grad[i] = (dbi_object.loss(s, d_new) - dbi_object.loss(s, d)) / delta
    return grad


def gradient_descent(
    dbi_object: DoubleBracketIteration,
    iterations: int,
    d_params: list,
    parameterization: ParameterizationTypes,
    pauli_operator_dict: dict = None,
    pauli_parameterization_order: int = 1,
    normalize: bool = False,
    lr_min: float = 1e-5,
    lr_max: float = 1,
    max_evals: int = 100,
    space: callable = None,
    optimizer: callable = hyperopt.tpe,
    verbose: bool = False,
    backend=None,
):
    r"""Numerical gradient descent method for variating diagonal operator in each double bracket rotation.

    Args:
        dbi_object (DoubleBracketIteration): the target double bracket object.
        iterations (int): number of double bracket rotations.
        d_params (list): the parameters for the initial diagonal operator.
        parameterization (ParameterizationTypes): the parameterization method for diagonal operator.
            Options include pauli and computational.
        pauli_operator_dict (dictionary, optional): dictionary of "name": Pauli-operator for Pauli-based parameterization type.
            Defaults to None.
        pauli_parameterization_order (int, optional): the order of parameterization or locality in Pauli basis. Defaults to 1.
        normalize (bool, optional): option to normalize the diagonal operator. Defaults to False.
        lr_min (float, optional): the minimal gradient step. Defaults to 1e-5.
        lr_max (float, optional): the maximal gradient step. Defaults to 1.
        max_evals (int, optional): maximum number of evaluations for `lr` using `hyperopt`. Defaults to 100.
        space (callable, optional): evalutation space for `hyperopt`. Defaults to None.
        optimizer (callable, optional): optimizer option for `hyperopt`. Defaults to `hyperopt.tpe`.
        verbose (bool, optional): option for printing `hyperopt` process. Defaults to False.

    Returns:
        loss_hist (list): list of history losses of `dbi_object` throughout the double bracket rotations.
        d_params_hist (list): list of history of `d` parameters after gradient descent.
        s_hist (list): list of history of optimal `s` found.
    Example:
        from qibo import set_backend
        from qibo.hamiltonians import Hamiltonian
        from qibo.models.dbi.double_bracket import *
        from qibo.models.dbi.utils import *
        from qibo.models.dbi.utils_dbr_strategies import gradient_descent
        from qibo.quantum_info import random_hermitian

        nqubits = 3
        NSTEPS = 5
        set_backend("numpy")
        h0 = random_hermitian(2**nqubits)
        dbi = DoubleBracketIteration(
            Hamiltonian(nqubits, h0),
            mode=DoubleBracketGeneratorType.single_commutator,
            scheduling=DoubleBracketScheduling.hyperopt,
            cost=DoubleBracketCostFunction.off_diagonal_norm,
        )
        initial_off_diagonal_norm = dbi.off_diagonal_norm
        pauli_operator_dict = generate_pauli_operator_dict(
            nqubits, parameterization_order=1
        )
        pauli_operators = list(pauli_operator_dict.values())
        # let initial d be approximation of $\Delta(H)
        d_coef_pauli = decompose_into_Pauli_basis(
            dbi.diagonal_h_matrix, pauli_operators=pauli_operators
        )
        d_pauli = sum([d_coef_pauli[i] * pauli_operators[i] for i in range(nqubits)])
        loss_hist_pauli, d_params_hist_pauli, s_hist_pauli = gradient_descent(
            dbi,
            NSTEPS,
            d_coef_pauli,
            ParameterizationTypes.pauli,
            pauli_operator_dict=pauli_operator_dict,
        )
    """
    backend = _check_backend(backend)

    nqubits = dbi_object.nqubits
    # TODO: write tests where this condition applies
    if (
        parameterization is ParameterizationTypes.pauli and pauli_operator_dict is None
    ):  # pragma: no cover
        pauli_operator_dict = generate_pauli_operator_dict(
            nqubits=nqubits, parameterization_order=pauli_parameterization_order
        )
    d = params_to_diagonal_operator(
        d_params,
        nqubits,
        parameterization=parameterization,
        pauli_operator_dict=pauli_operator_dict,
        normalize=normalize,
    )
    loss_hist = [dbi_object.loss(0.0, d=d)]
    d_params_hist = [d_params]
    s_hist = [0]
    # TODO: write tests where this condition applies
    if (
        parameterization is ParameterizationTypes.pauli and pauli_operator_dict is None
    ):  # pragma: no cover
        pauli_operator_dict = generate_pauli_operator_dict(
            nqubits=nqubits,
            parameterization_order=pauli_parameterization_order,
            backend=backend,
        )
    # first step
    s = dbi_object.choose_step(d=d)
    dbi_object(step=s, d=d)
    for _ in range(iterations):
        grad = gradient_numerical(
            dbi_object,
            d_params,
            parameterization,
            pauli_operator_dict=pauli_operator_dict,
            pauli_parameterization_order=pauli_parameterization_order,
            normalize=normalize,
            backend=backend,
        )

        # set up hyperopt to find optimal lr
        def func_loss_to_lr(lr):
            d_params_eval = [d_params[j] - grad[j] * lr for j in range(len(grad))]
            d_eval = params_to_diagonal_operator(
                d_params_eval,
                nqubits,
                parameterization=parameterization,
                pauli_operator_dict=pauli_operator_dict,
                normalize=normalize,
            )
            return dbi_object.loss(step=s, d=d_eval)

        if space is None:
            space = hyperopt.hp.loguniform("lr", np.log(lr_min), np.log(lr_max))

        best = hyperopt.fmin(
            fn=func_loss_to_lr,
            space=space,
            algo=optimizer.suggest,
            max_evals=max_evals,
            verbose=verbose,
        )
        lr = best["lr"]

        d_params = [d_params[j] - grad[j] * lr for j in range(len(grad))]
        d = params_to_diagonal_operator(
            d_params,
            nqubits,
            parameterization=parameterization,
            pauli_operator_dict=pauli_operator_dict,
            normalize=normalize,
        )
        s = dbi_object.choose_step(d=d)
        dbi_object(step=s, d=d)

        # record history
        loss_hist.append(dbi_object.loss(0.0, d=d))
        d_params_hist.append(d_params)
        s_hist.append(s)
    return loss_hist, d_params_hist, s_hist



import collections
import warnings
from typing import Optional, Union

import numpy as np

from qibo import __version__, backends, gates
from qibo.config import raise_error
from qibo.measurements import apply_bitflips, frequencies_to_binary


def load_result(filename: str):
    """Loads the results of a circuit execution saved to disk.

    Args:
        filename (str): Path to the file containing the results.

    Returns:
        :class:`qibo.result.QuantumState` or :class:`qibo.result.MeasurementOutcomes` or :class:`qibo.result.CircuitResult`: result of circuit execution saved to disk, depending on saved filed.
    """
    payload = np.load(filename, allow_pickle=True).item()
    return globals()[payload.pop("dtype")].from_dict(payload)


class QuantumState:
    """Data structure to represent the final state after circuit execution.

    Args:
        state (np.ndarray): Input quantum state as np.ndarray.
        backend (qibo.backends.AbstractBackend): Backend used for the calculations. If not provided the :class:`qibo.backends.GlobalBackend` is going to be used.
    """

    def __init__(self, state, backend=None):
        from qibo.backends import _check_backend

        self.backend = _check_backend(backend)
        self.density_matrix = len(state.shape) == 2
        self.nqubits = int(np.log2(state.shape[0]))
        self._state = state

    def symbolic(self, decimals: int = 5, cutoff: float = 1e-10, max_terms: int = 20):
        """Dirac notation representation of the state in the computational basis.

        Args:
            decimals (int, optional): Number of decimals for the amplitudes.
                Defaults to :math:`5`.
            cutoff (float, optional): Amplitudes with absolute value smaller than the
                cutoff are ignored from the representation. Defaults to  ``1e-10``.
            max_terms (int, optional): Maximum number of terms to print. If the state
                contains more terms they will be ignored. Defaults to :math:`20`.

        Returns:
            (str): A string representing the state in the computational basis.
        """
        if self.density_matrix:
            terms = self.backend.calculate_symbolic_density_matrix(
                self._state, self.nqubits, decimals, cutoff, max_terms
            )
        else:
            terms = self.backend.calculate_symbolic(
                self._state, self.nqubits, decimals, cutoff, max_terms
            )
        return " + ".join(terms)

    def state(self, numpy: bool = False):
        """State's tensor representation as a backend tensor.

        Args:
            numpy (bool, optional): If ``True`` the returned tensor will be a ``numpy`` array,
                otherwise it will follow the backend tensor type.
                Defaults to ``False``.

        Returns:
            The state in the computational basis.
        """
        if numpy:
            return np.array(self._state.tolist())

        return self._state

    def probabilities(self, qubits: Optional[Union[list, set]] = None):
        """Calculates measurement probabilities by tracing out qubits.

        When noisy model is applied to a circuit and `circuit.density_matrix=False`,
        this method returns the average probability resulting from
        repeated execution. This probability distribution approximates the
        exact probability distribution obtained when `circuit.density_matrix=True`.

        Args:
            qubits (list or set, optional): Set of qubits that are measured.
                If ``None``, ``qubits`` equates the total number of qubits.
                Defauts to ``None``.
        Returns:
            (np.ndarray): Probabilities over the input qubits.
        """

        if qubits is None:
            qubits = tuple(range(self.nqubits))

        if self.density_matrix:
            return self.backend.calculate_probabilities_density_matrix(
                self._state, qubits, self.nqubits
            )

        return self.backend.calculate_probabilities(self._state, qubits, self.nqubits)

    def __str__(self):
        return self.symbolic()

    def to_dict(self):
        """Returns a dictonary containinig all the information needed to rebuild the ``QuantumState``"""
        return {
            "state": self.state(numpy=True),
            "dtype": self.__class__.__name__,
            "qibo": __version__,
        }

    def dump(self, filename: str):
        """Writes to file the ``QuantumState`` for future reloading.

        Args:
            filename (str): Path to the file to write to.
        """
        with open(filename, "wb") as f:
            np.save(f, self.to_dict())

    @classmethod
    def from_dict(cls, payload: dict):
        """Builds a ``QuantumState`` object starting from a dictionary.

        Args:
            payload (dict): Dictionary containing all the information
                to load the ``QuantumState`` object.

        Returns:
            :class:`qibo.result.QuantumState`: Quantum state object..
        """
        backend = backends.construct_backend("numpy")
        return cls(payload.get("state"), backend=backend)

    @classmethod
    def load(cls, filename: str):
        """Builds the ``QuantumState`` object stored in a file.

        Args:
            filename (str): Path to the file containing the ``QuantumState``.

        Returns:
            :class:`qibo.result.QuantumState`: Quantum state object.
        """
        payload = np.load(filename, allow_pickle=True).item()
        return cls.from_dict(payload)


class MeasurementOutcomes:
    """Object to store the outcomes of measurements after circuit execution.

    Args:
        measurements (:class:`qibo.gates.M`): Measurement gates.
        backend (:class:`qibo.backends.AbstractBackend`): Backend used for the calculations.
            If ``None``, then :class:`qibo.backends.GlobalBackend` is used. Defaults to ``None``.
        probabilities (np.ndarray): Use these probabilities to generate samples and frequencies.
        samples (np.darray): Use these samples to generate probabilities and frequencies.
        nshots (int): Number of shots used for samples, probabilities and frequencies generation.
    """

    def __init__(
        self,
        measurements,
        backend=None,
        probabilities=None,
        samples: Optional[int] = None,
        nshots: int = 1000,
    ):
        self.backend = backend
        self.measurements = measurements
        self.nshots = nshots

        self._measurement_gate = None
        self._probs = probabilities
        self._samples = samples
        self._frequencies = None
        self._repeated_execution_frequencies = None

        if samples is not None:
            for m in measurements:
                indices = [self.measurement_gate.qubits.index(q) for q in m.qubits]
                m.result.register_samples(samples[:, indices])
        else:
            for gate in self.measurements:
                gate.result.reset()

    def frequencies(self, binary: bool = True, registers: bool = False):
        """Returns the frequencies of measured samples.

        Args:
            binary (bool, optional): Return frequency keys in binary or decimal form.
            registers (bool, optional): Group frequencies according to registers.

        Returns:
            A `collections.Counter` where the keys are the observed values
            and the values the corresponding frequencies, that is the number
            of times each measured value/bitstring appears.

            If ``binary`` is ``True``
                the keys of the `Counter` are in binary form, as strings of
                :math:`0`s and :math`1`s.
            If ``binary`` is ``False``
                the keys of the ``Counter`` are integers.
            If ``registers`` is ``True``
                a `dict` of `Counter` s is returned where keys are the name of
                each register.
            If ``registers`` is ``False``
                a single ``Counter`` is returned which contains samples from all
                the measured qubits, independently of their registers.
        """
        qubits = self.measurement_gate.qubits

        if self._repeated_execution_frequencies is not None:
            if binary:
                return self._repeated_execution_frequencies

            return collections.Counter(
                {int(k, 2): v for k, v in self._repeated_execution_frequencies.items()}
            )

        if self._frequencies is None:
            if self.measurement_gate.has_bitflip_noise() and not self.has_samples():
                self._samples = self.samples()
            if not self.has_samples():
                # generate new frequencies
                self._frequencies = self.backend.sample_frequencies(
                    self._probs, self.nshots
                )
                # register frequencies to individual gate ``MeasurementResult``
                qubit_map = {q: i for i, q in enumerate(qubits)}
                reg_frequencies = {}
                binary_frequencies = frequencies_to_binary(
                    self._frequencies, len(qubits)
                )
                for gate in self.measurements:
                    rfreqs = collections.Counter()
                    for bitstring, freq in binary_frequencies.items():
                        idx = 0
                        rqubits = gate.target_qubits
                        for i, q in enumerate(rqubits):
                            if int(bitstring[qubit_map.get(q)]):
                                idx += 2 ** (len(rqubits) - i - 1)
                        rfreqs[idx] += freq
                    gate.result.register_frequencies(rfreqs, self.backend)
            else:
                self._frequencies = self.backend.calculate_frequencies(
                    self.samples(binary=False)
                )

        if registers:
            return {
                gate.register_name: gate.result.frequencies(binary)
                for gate in self.measurements
            }

        if binary:
            return frequencies_to_binary(self._frequencies, len(qubits))

        return self._frequencies

    def probabilities(self, qubits: Optional[Union[list, set]] = None):
        """Calculate the probabilities as frequencies / nshots

        Returns:
            The array containing the probabilities of the measured qubits.
        """
        nqubits = len(self.measurement_gate.qubits)
        if qubits is None:
            qubits = range(nqubits)
        else:
            if not set(qubits).issubset(self.measurement_gate.qubits):
                raise_error(
                    RuntimeError,
                    f"Asking probabilities for qubits {qubits}, but only qubits {self.measurement_gate.qubits} were measured.",
                )
            qubits = [self.measurement_gate.qubits.index(q) for q in qubits]

        if self._probs is not None and not self.measurement_gate.has_bitflip_noise():
            return self.backend.calculate_probabilities(
                np.sqrt(self._probs), qubits, nqubits
            )

        probs = [0 for _ in range(2**nqubits)]
        for state, freq in self.frequencies(binary=False).items():
            probs[state] = freq / self.nshots
        probs = self.backend.cast(probs)
        self._probs = probs
        return self.backend.calculate_probabilities(np.sqrt(probs), qubits, nqubits)

    def has_samples(self):
        """Check whether the samples are available already.

        Returns:
            (bool): ``True`` if the samples are available, ``False`` otherwise.
        """
        return self.measurements[0].result.has_samples() or self._samples is not None

    def samples(self, binary: bool = True, registers: bool = False):
        """Returns raw measurement samples.

        Args:
            binary (bool, optional): Return samples in binary or decimal form.
            registers (bool, optional): Group samples according to registers.

        Returns:
            If ``binary`` is ``True``
                samples are returned in binary form as a tensor
                of shape ``(nshots, n_measured_qubits)``.
            If ``binary`` is ``False``
                samples are returned in decimal form as a tensor
                of shape ``(nshots,)``.
            If ``registers`` is ``True``
                samples are returned in a ``dict`` where the keys are the register
                names and the values are the samples tensors for each register.
            If ``registers`` is ``False``
                a single tensor is returned which contains samples from all the
                measured qubits, independently of their registers.
        """
        qubits = self.measurement_gate.target_qubits
        if self._samples is None:
            if self.measurements[0].result.has_samples():
                self._samples = self.backend.np.concatenate(
                    [gate.result.samples() for gate in self.measurements], axis=1
                )
            else:
                if self._frequencies is not None:
                    # generate samples that respect the existing frequencies
                    frequencies = self.frequencies(binary=False)
                    samples = np.concatenate(
                        [np.repeat(x, f) for x, f in frequencies.items()]
                    )
                    np.random.shuffle(samples)
                else:
                    # generate new samples
                    samples = self.backend.sample_shots(self._probs, self.nshots)
                samples = self.backend.samples_to_binary(samples, len(qubits))
                if self.measurement_gate.has_bitflip_noise():
                    p0, p1 = self.measurement_gate.bitflip_map
                    bitflip_probabilities = [
                        [p0.get(q) for q in qubits],
                        [p1.get(q) for q in qubits],
                    ]
                    samples = self.backend.apply_bitflips(
                        samples, bitflip_probabilities
                    )
                # register samples to individual gate ``MeasurementResult``
                qubit_map = {
                    q: i for i, q in enumerate(self.measurement_gate.target_qubits)
                }
                self._samples = self.backend.cast(samples, "int32")
                for gate in self.measurements:
                    rqubits = tuple(qubit_map.get(q) for q in gate.target_qubits)
                    gate.result.register_samples(
                        self._samples[:, rqubits], self.backend
                    )

        if registers:
            return {
                gate.register_name: gate.result.samples(binary)
                for gate in self.measurements
            }

        if binary:
            return self._samples

        return self.backend.samples_to_decimal(self._samples, len(qubits))

    @property
    def measurement_gate(self):
        """Single measurement gate containing all measured qubits.

        Useful for sampling all measured qubits at once when simulating.
        """
        if self._measurement_gate is None:
            for gate in self.measurements:
                if self._measurement_gate is None:
                    self._measurement_gate = gates.M(
                        *gate.init_args, **gate.init_kwargs
                    )
                else:
                    self._measurement_gate.add(gate)

        return self._measurement_gate

    def apply_bitflips(self, p0: float, p1: Optional[float] = None):
        """Apply bitflips to the measurements with probabilities `p0` and `p1`

        Args:
            p0 (float): Probability of the 0->1 flip.
            p1 (float): Probability of the 1->0 flip.
        """
        return apply_bitflips(self, p0, p1)

    def expectation_from_samples(self, observable):
        """Computes the real expectation value of a diagonal observable from frequencies.

        Args:
            observable (Hamiltonian/SymbolicHamiltonian): diagonal observable in the
                computational basis.

        Returns:
            (float): expectation value from samples.
        """
        freq = self.frequencies(binary=True)
        qubit_map = self.measurement_gate.qubits
        return observable.expectation_from_samples(freq, qubit_map)

    def to_dict(self):
        """Returns a dictonary containinig all the information needed to rebuild the :class:`qibo.result.MeasurementOutcomes`."""
        args = {
            "measurements": [m.to_json() for m in self.measurements],
            "probabilities": self._probs,
            "samples": self._samples,
            "nshots": self.nshots,
            "dtype": self.__class__.__name__,
            "qibo": __version__,
        }
        return args

    def dump(self, filename: str):
        """Writes to file the :class:`qibo.result.MeasurementOutcomes` for future reloading.

        Args:
            filename (str): Path to the file to write to.
        """
        with open(filename, "wb") as f:
            np.save(f, self.to_dict())

    @classmethod
    def from_dict(cls, payload: dict):
        """Builds a :class:`qibo.result.MeasurementOutcomes` object starting from a dictionary.

        Args:
            payload (dict): Dictionary containing all the information to load the :class:`qibo.result.MeasurementOutcomes` object.

        Returns:
            A :class:`qibo.result.MeasurementOutcomes` object.
        """
        from qibo.backends import construct_backend

        if payload["probabilities"] is not None and payload["samples"] is not None:
            warnings.warn(
                "Both `probabilities` and `samples` found, discarding the `probabilities` and building out of the `samples`."
            )
            payload.pop("probabilities")
        backend = construct_backend("numpy")
        measurements = [gates.M.load(m) for m in payload.get("measurements")]
        return cls(
            measurements,
            backend=backend,
            probabilities=payload.get("probabilities"),
            samples=payload.get("samples"),
            nshots=payload.get("nshots"),
        )

    @classmethod
    def load(cls, filename: str):
        """Builds the :class:`qibo.result.MeasurementOutcomes` object stored in a file.

        Args:
            filename (str): Path to the file containing the :class:`qibo.result.MeasurementOutcomes`.

        Returns:
            A :class:`qibo.result.MeasurementOutcomes` object.
        """
        payload = np.load(filename, allow_pickle=True).item()
        return cls.from_dict(payload)


class CircuitResult(QuantumState, MeasurementOutcomes):
    """Object to store both the outcomes of measurements and the final state after circuit execution.

    Args:
        final_state (np.ndarray): Input quantum state as np.ndarray.
        measurements (qibo.gates.M): The measurement gates containing the measurements.
        backend (qibo.backends.AbstractBackend): Backend used for the calculations. If not provided the :class:`qibo.backends.GlobalBackend` is going to be used.
        probabilities (np.ndarray): Use these probabilities to generate samples and frequencies.
        samples (np.darray): Use these samples to generate probabilities and frequencies.
        nshots (int): Number of shots used for samples, probabilities and frequencies generation.
    """

    def __init__(
        self, final_state, measurements, backend=None, samples=None, nshots=1000
    ):
        QuantumState.__init__(self, final_state, backend)
        qubits = [q for m in measurements for q in m.target_qubits]
        if len(qubits) == 0:
            raise ValueError(
                "Circuit does not contain measurements. Use a `QuantumState` instead."
            )
        probs = QuantumState.probabilities(self, qubits) if samples is None else None
        MeasurementOutcomes.__init__(
            self,
            measurements,
            backend=backend,
            probabilities=probs,
            samples=samples,
            nshots=nshots,
        )

    def probabilities(self, qubits: Optional[Union[list, set]] = None):
        if self.measurement_gate.has_bitflip_noise():
            return MeasurementOutcomes.probabilities(self, qubits)
        return QuantumState.probabilities(self, qubits)

    def to_dict(self):
        """Returns a dictonary containinig all the information needed to rebuild the ``CircuitResult``."""
        args = MeasurementOutcomes.to_dict(self)
        args.update(QuantumState.to_dict(self))
        args.update({"dtype": self.__class__.__name__})
        return args

    @classmethod
    def from_dict(cls, payload: dict):
        """Builds a ``CircuitResult`` object starting from a dictionary.

        Args:
            payload (dict): Dictionary containing all the information to load the ``CircuitResult`` object.

        Returns:
            :class:`qibo.result.CircuitResult`: circuit result object.
        """
        state_load = {"state": payload.pop("state")}
        state = QuantumState.from_dict(state_load)
        measurements = MeasurementOutcomes.from_dict(payload)
        return cls(
            state.state(),
            measurements.measurements,
            backend=state.backend,
            samples=measurements.samples(),
            nshots=measurements.nshots,
        )



import os
from importlib import import_module

import numpy as np

from qibo.backends.abstract import Backend
from qibo.backends.clifford import CliffordBackend
from qibo.backends.npmatrices import NumpyMatrices
from qibo.backends.numpy import NumpyBackend
from qibo.backends.pytorch import PyTorchBackend
from qibo.backends.tensorflow import TensorflowBackend
from qibo.config import log, raise_error

QIBO_NATIVE_BACKENDS = ("numpy", "tensorflow", "pytorch")
QIBO_NON_NATIVE_BACKENDS = ("qibojit", "qibolab", "qibo-cloud-backends", "qibotn")


class MetaBackend:
    """Meta-backend class which takes care of loading the qibo backends."""

    @staticmethod
    def load(backend: str, **kwargs) -> Backend:
        """Loads the native qibo backend.

        Args:
            backend (str): Name of the backend to load.
            kwargs (dict): Additional arguments for the qibo backend.
        Returns:
            qibo.backends.abstract.Backend: The loaded backend.
        """

        if backend == "numpy":
            return NumpyBackend()
        elif backend == "tensorflow":
            return TensorflowBackend()
        elif backend == "pytorch":
            return PyTorchBackend()
        elif backend == "clifford":
            engine = kwargs.pop("platform", None)
            kwargs["engine"] = engine
            return CliffordBackend(**kwargs)
        else:
            raise_error(
                ValueError,
                f"Backend {backend} is not available. The native qibo backends are {QIBO_NATIVE_BACKENDS}.",
            )

    def list_available(self) -> dict:
        """Lists all the available native qibo backends."""
        available_backends = {}
        for backend in QIBO_NATIVE_BACKENDS:
            try:
                MetaBackend.load(backend)
                available = True
            except:  # pragma: no cover
                available = False
            available_backends[backend] = available
        return available_backends


class GlobalBackend(NumpyBackend):
    """The global backend will be used as default by ``circuit.execute()``."""

    _instance = None
    _dtypes = {"double": "complex128", "single": "complex64"}
    _default_order = [
        {"backend": "qibojit", "platform": "cupy"},
        {"backend": "qibojit", "platform": "numba"},
        {"backend": "tensorflow"},
        {"backend": "numpy"},
        {"backend": "pytorch"},
    ]

    def __new__(cls):
        if cls._instance is not None:
            return cls._instance

        backend = os.environ.get("QIBO_BACKEND")
        if backend:  # pragma: no cover
            # Create backend specified by user
            platform = os.environ.get("QIBO_PLATFORM")
            cls._instance = construct_backend(backend, platform=platform)
        else:
            # Create backend according to default order
            for kwargs in cls._default_order:
                try:
                    cls._instance = construct_backend(**kwargs)
                    break
                except (ModuleNotFoundError, ImportError):
                    pass

        if cls._instance is None:  # pragma: no cover
            raise_error(RuntimeError, "No backends available.")

        log.info(f"Using {cls._instance} backend on {cls._instance.device}")
        return cls._instance

    @classmethod
    def set_backend(cls, backend, **kwargs):  # pragma: no cover
        if (
            cls._instance is None
            or cls._instance.name != backend
            or cls._instance.platform != kwargs.get("platform")
        ):
            cls._instance = construct_backend(backend, **kwargs)
        log.info(f"Using {cls._instance} backend on {cls._instance.device}")


class QiboMatrices:
    def __init__(self, dtype="complex128"):
        self.create(dtype)

    def create(self, dtype):
        self.matrices = NumpyMatrices(dtype)
        self.I = self.matrices.I(2)
        self.X = self.matrices.X
        self.Y = self.matrices.Y
        self.Z = self.matrices.Z
        self.SX = self.matrices.SX
        self.H = self.matrices.H
        self.S = self.matrices.S
        self.SDG = self.matrices.SDG
        self.CNOT = self.matrices.CNOT
        self.CY = self.matrices.CY
        self.CZ = self.matrices.CZ
        self.CSX = self.matrices.CSX
        self.CSXDG = self.matrices.CSXDG
        self.SWAP = self.matrices.SWAP
        self.iSWAP = self.matrices.iSWAP
        self.SiSWAP = self.matrices.SiSWAP
        self.SiSWAPDG = self.matrices.SiSWAPDG
        self.FSWAP = self.matrices.FSWAP
        self.ECR = self.matrices.ECR
        self.SYC = self.matrices.SYC
        self.TOFFOLI = self.matrices.TOFFOLI
        self.CCZ = self.matrices.CCZ


matrices = QiboMatrices()


def get_backend():
    return str(GlobalBackend())


def set_backend(backend, **kwargs):
    GlobalBackend.set_backend(backend, **kwargs)


def get_precision():
    return GlobalBackend().precision


def set_precision(precision):
    GlobalBackend().set_precision(precision)
    matrices.create(GlobalBackend().dtype)


def get_device():
    return GlobalBackend().device


def set_device(device):
    parts = device[1:].split(":")
    if device[0] != "/" or len(parts) < 2 or len(parts) > 3:
        raise_error(
            ValueError,
            "Device name should follow the pattern: /{device type}:{device number}.",
        )
    backend = GlobalBackend()
    backend.set_device(device)
    log.info(f"Using {backend} backend on {backend.device}")


def get_threads():
    return GlobalBackend().nthreads


def set_threads(nthreads):
    if not isinstance(nthreads, int):
        raise_error(TypeError, "Number of threads must be integer.")
    if nthreads < 1:
        raise_error(ValueError, "Number of threads must be positive.")
    GlobalBackend().set_threads(nthreads)


def _check_backend(backend):
    if backend is None:
        return GlobalBackend()

    return backend


def list_available_backends() -> dict:
    """Lists all the backends that are available."""
    available_backends = MetaBackend().list_available()
    for backend in QIBO_NON_NATIVE_BACKENDS:
        try:
            module = import_module(backend.replace("-", "_"))
            available = getattr(module, "MetaBackend")().list_available()
        except:
            available = False
        available_backends.update({backend: available})
    return available_backends


def construct_backend(backend, **kwargs) -> Backend:
    """Construct a generic native or non-native qibo backend.
    Args:
        backend (str): Name of the backend to load.
        kwargs (dict): Additional arguments for constructing the backend.
    Returns:
        qibo.backends.abstract.Backend: The loaded backend.

    """
    if backend in QIBO_NATIVE_BACKENDS + ("clifford",):
        return MetaBackend.load(backend, **kwargs)
    elif backend in QIBO_NON_NATIVE_BACKENDS:
        module = import_module(backend.replace("-", "_"))
        return getattr(module, "MetaBackend").load(**kwargs)
    else:
        raise_error(
            ValueError,
            f"Backend {backend} is not available. To check which backends are installed use `qibo.list_available_backends()`.",
        )


def _check_backend_and_local_state(seed, backend):
    if (
        seed is not None
        and not isinstance(seed, int)
        and not isinstance(seed, np.random.Generator)
    ):
        raise_error(
            TypeError, "seed must be either type int or numpy.random.Generator."
        )

    backend = _check_backend(backend)

    if seed is None or isinstance(seed, int):
        if backend.__class__.__name__ in [
            "CupyBackend",
            "CuQuantumBackend",
        ]:  # pragma: no cover
            local_state = backend.np.random.default_rng(seed)
        else:
            local_state = np.random.default_rng(seed)
    else:
        local_state = seed

    return backend, local_state



import math
from enum import Enum, auto
from itertools import combinations, product

import numpy as np

from qibo import symbols
from qibo.backends import _check_backend
from qibo.hamiltonians import SymbolicHamiltonian


def generate_Z_operators(nqubits: int, backend=None):
    """Generate a dictionary containing 1) all possible products of Pauli Z operators for L = n_qubits and 2) their respective names.
    Return: Dictionary with operator names (str) as keys and operators (np.array) as values

     Example:
        .. testcode::

            from qibo.models.dbi.utils import generate_Z_operators
            from qibo.models.dbi.double_bracket import DoubleBracketIteration
            from qibo.quantum_info import random_hermitian
            from qibo.hamiltonians import Hamiltonian
            import numpy as np

            nqubits = 4
            h0 = random_hermitian(2**nqubits)
            dbi = DoubleBracketIteration(Hamiltonian(nqubits=nqubits, matrix=h0))
            generate_Z = generate_Z_operators(nqubits)
            Z_ops = list(generate_Z.values())

            delta_h0 = dbi.diagonal_h_matrix
            dephasing_channel = (sum([Z_op @ h0 @ Z_op for Z_op in Z_ops])+h0)/2**nqubits
            norm_diff = np.linalg.norm(delta_h0 - dephasing_channel)
    """

    backend = _check_backend(backend)
    # list of tuples, e.g. ('Z','I','Z')
    combination_strings = product("ZI", repeat=nqubits)
    output_dict = {}

    for zi_string_combination in combination_strings:
        # except for the identity
        if "Z" in zi_string_combination:
            op_name = "".join(zi_string_combination)
            tensor_op = str_to_symbolic(op_name)
            # append in output_dict
            output_dict[op_name] = SymbolicHamiltonian(
                tensor_op, backend=backend
            ).dense.matrix
    return output_dict


def str_to_symbolic(name: str):
    """Convert string into symbolic hamiltonian.
    Example:
        .. testcode::

            from qibo.models.dbi.utils import str_to_symbolic
            op_name = "ZYXZI"
            # returns 5-qubit symbolic hamiltonian
            ZIXZI_op = str_to_symbolic(op_name)
    """
    tensor_op = 1
    for qubit, char in enumerate(name):
        tensor_op *= getattr(symbols, char)(qubit)
    return tensor_op


def cs_angle_sgn(dbi_object, d):
    """Calculates the sign of Cauchy-Schwarz Angle :math:`\\langle W(Z), W({\\rm canonical}) \\rangle_{\\rm HS}`."""
    d = dbi_object.backend.cast(d)
    norm = np.trace(
        np.dot(
            np.conjugate(
                dbi_object.commutator(dbi_object.diagonal_h_matrix, dbi_object.h.matrix)
            ).T,
            dbi_object.commutator(d, dbi_object.h.matrix),
        )
    )
    return np.real(np.sign(norm))


def decompose_into_pauli_basis(h_matrix: np.array, pauli_operators: list):
    """finds the decomposition of hamiltonian `h_matrix` into Pauli-Z operators"""
    nqubits = int(np.log2(h_matrix.shape[0]))

    decomposition = []
    for Z_i in pauli_operators:
        expect = np.trace(h_matrix @ Z_i) / 2**nqubits
        decomposition.append(expect)
    return decomposition


def generate_pauli_index(nqubits, order):
    """
    Generate all possible combinations of qubits for a given order of Pauli operators.
    """
    if order == 1:
        return list(range(nqubits))
    else:
        indices = list(range(nqubits))
        return indices + [
            comb for i in range(2, order + 1) for comb in combinations(indices, i)
        ]


def generate_pauli_operator_dict(
    nqubits: int,
    parameterization_order: int = 1,
    symbols_pauli=symbols.Z,
    backend=None,
):
    """Generates a dictionary containing Pauli `symbols_pauli` operators of locality `parameterization_order` for `nqubits` qubits.

    Args:
        nqubits (int): number of qubits in the system.
        parameterization_order (int, optional): the locality of the operators generated. Defaults to 1.
        symbols_pauli (qibo.symbols, optional): the symbol of the intended Pauli operator. Defaults to symbols.Z.

    Returns:
        pauli_operator_dict (dictionary): dictionary with structure {"operator_name": operator}

    Example:
        pauli_operator_dict = generate_pauli_operator_dict)
    """
    backend = _check_backend(backend)
    pauli_index = generate_pauli_index(nqubits, order=parameterization_order)
    pauli_operators = [
        generate_pauli_operators(nqubits, symbols_pauli, index, backend=backend)
        for index in pauli_index
    ]
    return {index: operator for index, operator in zip(pauli_index, pauli_operators)}


def generate_pauli_operators(nqubits, symbols_pauli, positions, backend=None):
    # generate matrix of an nqubit-pauli operator with `symbols_pauli` at `positions`
    if isinstance(positions, int):
        return SymbolicHamiltonian(
            symbols_pauli(positions),
            nqubits=nqubits,
            backend=backend,
        ).dense.matrix
    else:
        terms = [symbols_pauli(pos) for pos in positions]
        return SymbolicHamiltonian(
            math.prod(terms), nqubits=nqubits, backend=backend
        ).dense.matrix


class ParameterizationTypes(Enum):
    """Define types of parameterization for diagonal operator."""

    pauli = auto()
    """Uses Pauli-Z operators (magnetic field)."""
    computational = auto()
    """Uses computational basis."""


def params_to_diagonal_operator(
    params: np.array,
    nqubits: int,
    parameterization: ParameterizationTypes = ParameterizationTypes.pauli,
    pauli_parameterization_order: int = 1,
    normalize: bool = False,
    pauli_operator_dict: dict = None,
    backend=None,
):
    r"""Creates the $D$ operator for the double-bracket iteration ansatz depending on the parameterization type."""
    backend = _check_backend(backend)
    if parameterization is ParameterizationTypes.pauli:
        # raise error if dimension mismatch
        d = sum(
            [
                backend.to_numpy(params[i])
                * backend.to_numpy(list(pauli_operator_dict.values())[i])
                for i in range(nqubits)
            ]
        )
    elif parameterization is ParameterizationTypes.computational:
        d = np.zeros((len(params), len(params)))
        for i in range(len(params)):
            d[i, i] = backend.to_numpy(params[i])

    # TODO: write proper tests for normalize=True
    if normalize:  # pragma: no cover
        d = d / np.linalg.norm(d)
    return d


def off_diagonal_norm_polynomial_expansion_coef(dbi_object, d, n):
    # generate Gamma's where $\Gamma_{k+1}=[W, \Gamma_{k}], $\Gamma_0=H
    W = dbi_object.commutator(
        dbi_object.backend.cast(d), dbi_object.sigma(dbi_object.h.matrix)
    )
    gamma_list = dbi_object.generate_gamma_list(n + 2, d)
    sigma_gamma_list = list(map(dbi_object.sigma, gamma_list))
    gamma_list_np = list(map(dbi_object.backend.to_numpy, sigma_gamma_list))
    exp_list = np.array([1 / math.factorial(k) for k in range(n + 1)])
    # coefficients for rotation with [W,H] and H
    c1 = exp_list.reshape((-1, 1, 1)) * gamma_list_np[1:]
    c2 = exp_list.reshape((-1, 1, 1)) * gamma_list_np[:-1]
    # product coefficient
    trace_coefficients = [0] * (2 * n + 1)
    for k in range(n + 1):
        for j in range(n + 1):
            power = k + j
            product_matrix = c1[k] @ c2[j]
            trace_coefficients[power] += 2 * np.trace(product_matrix)
    # coefficients from high to low (n:0)
    coef = list(reversed(trace_coefficients[: n + 1]))
    return coef


def least_squares_polynomial_expansion_coef(dbi_object, d, n: int = 3):
    """Return the Taylor expansion coefficients of least square cost of `dbi_object.h` and diagonal operator `d` with respect to double bracket rotation duration `s`."""
    # generate Gamma's where $\Gamma_{k+1}=[W, \Gamma_{k}], $\Gamma_0=H
    Gamma_list = dbi_object.generate_gamma_list(n + 1, d)
    exp_list = np.array([1 / math.factorial(k) for k in range(n + 1)])
    # coefficients
    coef = np.empty(n)
    for i in range(n):
        coef[i] = np.real(
            exp_list[i] * np.trace(dbi_object.backend.cast(d) @ Gamma_list[i + 1])
        )
    coef = list(reversed(coef))
    return coef


def energy_fluctuation_polynomial_expansion_coef(
    dbi_object, d: np.array, n: int = 3, state=0
):
    """Return the Taylor expansion coefficients of energy fluctuation of `dbi_object` with respect to double bracket rotation duration `s`."""
    # generate Gamma's where $\Gamma_{k+1}=[W, \Gamma_{k}], $\Gamma_0=H
    Gamma_list = dbi_object.generate_gamma_list(n + 1, d)
    # coefficients
    coef = np.empty(3)
    state_cast = dbi_object.backend.cast(state)
    state_dag = dbi_object.backend.cast(state.conj().T)

    def variance(a):
        """Calculates the variance of a matrix A with respect to a state:
        Var($A$) = $\\langle\\mu|A^2|\\mu\rangle-\\langle\\mu|A|\\mu\rangle^2$"""
        b = a @ a
        return state_dag @ b @ state_cast - (state_dag @ a @ state_cast) ** 2

    def covariance(a, b):
        """This is a generalization of the notion of covariance, needed for the polynomial expansion of the energy fluctuation,
        applied to two operators A and B with respect to a state:
        Cov($A,B$) = $\\langle\\mu|AB|\\mu\rangle-\\langle\\mu|A|\\mu\rangle\\langle\\mu|B|\\mu\rangle$
        """

        c = a @ b + b @ a
        return (
            state_dag @ c @ state_cast
            - 2 * state_dag @ a @ state_cast * state_dag @ b @ state_cast
        )

    coef[0] = np.real(2 * covariance(Gamma_list[0], Gamma_list[1]))
    coef[1] = np.real(2 * variance(Gamma_list[1]))
    coef[2] = np.real(
        covariance(Gamma_list[0], Gamma_list[3])
        + 3 * covariance(Gamma_list[1], Gamma_list[2])
    )
    coef = list(reversed(coef))
    return coef
